from ewah.operators.sql_base_operator import EWAHSQLBaseOperator
from ewah.constants import EWAHConstants as EC

from airflow.hooks.base_hook import BaseHook

from google.cloud import bigquery
from tempfile import NamedTemporaryFile

import os

class EWAHBigQueryOperator(EWAHSQLBaseOperator):

    _SQL_BASE = \
        'SELECT\n{columns}\nFROM `{schema}`.`{table}`\nWHERE {where_clause}'
    _SQL_BASE_SELECT = \
        'SELECT * FROM ({select_sql}) t WHERE {{0}}'
    _SQL_COLUMN_QUOTE = '`'
    _SQL_MINMAX_CHUNKS = 'SELECT MIN({column}), MAX({column}) FROM ({base}) t;'
    _SQL_CHUNKING_CLAUSE = '''
        AND {column} >= @from_value
        AND {column} <{equal_sign} @until_value
    '''
    _SQL_PARAMS = '@{0}'

    def __init__(self, *args, **kwargs):
        self.sql_engine = self._BQ
        super().__init__(*args, **kwargs)

    def _get_data_from_sql(self, sql, params=None, return_dict=True):

        # set params
        if params:
            type_mapping = EC.QBC_TYPE_MAPPING.get(EC.DWH_ENGINE_BIGQUERY)
            _default = type_mapping[EC.QBC_TYPE_MAPPING_DEFAULT]
            job_config = bigquery.QueryJobConfig(
                query_parameters=[
                    bigquery.ScalarQueryParameter(
                        key,
                        type_mapping.get(type(value), _default),
                        value,
                    )
                    for key, value in params.items()
                ]
            )
        else:
            job_config = None

        # create temp file with creds
        self.log.info('Executing:\n{0}\n\nWith params:\n{1}'.format(
            sql,
            str(params),
        ))
        with NamedTemporaryFile() as cred_json:
            cred_filename = os.path.abspath(cred_json.name)
            cred_json.seek(0)
            cred_json.write(self.source_conn.extra.encode())
            cred_json.seek(0)
            cred_json.flush()
            client = bigquery.Client.from_service_account_json(cred_filename)
            results = client.query(sql, job_config=job_config).result()
            dict_result = [dict(row) for row in results]

        if return_dict or not dict_result:
            return dict_result
        keys = dict_result[0].keys()
        return [[row[key] for key in key] for row in dict_result]



from ewah.constants import EWAHConstants as EC
from ewah.dwhooks.dwhook_snowflake import EWAHDWHookSnowflake
from ewah.dwhooks.dwhook_postgres import EWAHDWHookPostgres
from ewah.dwhooks.dwhook_google_sheets import EWAHDWHookGSheets

def get_dwhook(dwh_engine):
    try:
        return {
            EC.DWH_ENGINE_POSTGRES: EWAHDWHookPostgres,
            EC.DWH_ENGINE_SNOWFLAKE: EWAHDWHookSnowflake,
            # DWH_ENGINE_BIGQUERY: bq_hook,
            EC.DWH_ENGINE_GS: EWAHDWHookGSheets,
        }[dwh_engine]
    except KeyError:
        raise Exception('Invalid Engine operator selected!')



from ewah.operators.sql_base_operator import EWAHSQLBaseOperator
from ewah.constants import EWAHConstants as EC

from airflow.hooks.base_hook import BaseHook

import pymysql

class EWAHMySQLOperator(EWAHSQLBaseOperator):

    _SQL_BASE = \
        'SELECT\n{columns}\nFROM `{schema}`.`{table}`\nWHERE {where_clause}'
    _SQL_BASE_SELECT = \
        'SELECT * FROM ({select_sql}) t WHERE {{0}}'
    _SQL_COLUMN_QUOTE = '`'
    _SQL_MINMAX_CHUNKS = 'SELECT MIN({column}), MAX({column}) FROM ({base}) t;'
    _SQL_CHUNKING_CLAUSE = '''
        AND {column} >= %(from_value)s
        AND {column} <{equal_sign} %(until_value)s
    '''
    _SQL_PARAMS = '%({0})s'

    def __init__(self, *args, **kwargs):
        self.sql_engine = self._MYSQL
        super().__init__(*args, **kwargs)

    def _get_data_from_sql(self, sql, params=None, return_dict=True):
        if return_dict:
            cursor_class = pymysql.cursors.DictCursor
        else:
            cursor_class = pymysql.cursors.Cursor
        database_conn = pymysql.connect(**{
            'host': self.source_conn.host,
            'user': self.source_conn.login,
            'passwd': self.source_conn.password,
            'port': self.source_conn.port,
            'database': self.source_conn.schema,
            'cursorclass': cursor_class,
        })
        cursor = database_conn.cursor()
        self.log.info('Executing:\n{0}\n\nWith params:\n{1}'.format(
            sql,
            str(params),
        ))
        cursor.execute(sql, args=params)
        data = cursor.fetchall()
        cursor.close()
        database_conn.close()
        return data



from ewah.dwhooks.base_dwhook import EWAHBaseDWHook
from ewah.constants import EWAHConstants as EC

from tempfile import NamedTemporaryFile
from datetime import datetime
from decimal import Decimal
import gspread
import json
import os

def monkeypatch_values_update(func_to_call):
    def monkeypatch_func(range_name, params, body):
        values = body.get('values')
        if values:
            values = [[ \
                float(v) if isinstance(v, Decimal)
                else (v.strftime('%Y-%m-%d %H:%M:%S%z') \
                    if isinstance(v, datetime)
                else v) for v in outer
            ] for outer in values]
            body['values'] = values
        return func_to_call(range_name, params=params, body=body)
    return monkeypatch_func

class EWAHDWHookGSheets(EWAHBaseDWHook):
    """
    The Google Sheets Hook works a little bit different to other hooks.
        - it can only have 1 upload call per task execution
        - it deletes all data from the sheet and re-uploads it to the sheet
        - the target schema name is the workbook key
        - the target table name is the worksheet name
    """

    def __init__(self, *args, **kwargs):
        self._upload_call_count = 0
        super().__init__(EC.DWH_ENGINE_GS, *args, **kwargs)

    def _init_conn(self, first_call=False, commit=False):
        # need to overwrite parent function, does nothing
        self.conn = None

    def execute(self, *args, **kwargs):
        # need to overwrite parent function, does nothing
        raise Exception('This function is not implemented for this DWHook!')

    def detect_and_apply_schema_changes(self, *args, **kwargs):
        # need to overwrite parent function, does nothing
        raise Exception('This function is not implemented for this DWHook!')

    def commit(self):
        self.logging_func('DWH Hook: Google Sheets is auto-commited')

    def rollback(self):
        raise Exception('Cannot rollback Google Sheets changes!')

    def _create_or_update_table(self,
        data,
        table_name,
        schema_name,
        schema_suffix, # unused but always given
        columns_definition, # unused but always given
        columns_partial_query, # unused but always given
        update_on_columns, # unused but always given
        drop_and_replace, # must be true!
        logging_function,
        pk_columns=[], # must accept arg, but it must also always be []
    ):
        if not pk_columns == []:
            raise Exception('Arg pk_columns invalidly supplied!')
            
        def colnum_string(n):
            # adapted from https://stackoverflow.com/questions/23861680/convert-spreadsheet-number-to-column-letter
            string = ""
            while n > 0:
                n, remainder = divmod(n - 1, 26)
                string = chr(65 + remainder) + string
            return string

        if not drop_and_replace:
            raise Exception('Google Sheets DWHs can only be drop_and_replace!')
        self._upload_call_count += 1
        if not self._upload_call_count == 1:
            raise Exception('Chunking is not possible for Google Sheets DWH!')
        logging_function('Replacing data in Google Sheets!')

        if not data:
            logging_function('Nothing to upload!')
            return

        # authorize and get correct worksheet
        servie_acc_file = NamedTemporaryFile()
        logging_function('Authenticating...')
        credentials = self.credentials.extra_dejson
        if credentials.get('client_secrets'):
            credentials = credentials['client_secrets']
        service_acc = json.dumps(credentials)
        filename = os.path.abspath(servie_acc_file.name)
        with open(filename, 'w') as f:
            f.write(service_acc)
        gclient = gspread.service_account(filename=filename)

        workbook = gclient.open_by_key(schema_name)
        worksheet = workbook.worksheet(table_name)
        values = worksheet.get_all_values()

        # Delete old data, if any existed
        if values:
            logging_function('Deleting old data...')
            rows = len(values)
            columns = len(values[0])
            values = [columns * ['']] * rows
            range_notation = 'A1:' + colnum_string(columns) + str(rows)
            worksheet.update(range_notation, values) # delete!

        del values # free up memory

        # insert new data - need to change the format! use columns definition
        logging_function('Preparing data for Google Sheets upload...')
        column_header = list(columns_definition.keys())
        upload_data = [column_header]
        while data:
            current_data = data.pop(0)
            upload_data+= [[current_data.get(col, '') for col in column_header]]
        range_notation = 'A1:' + colnum_string(len(column_header))
        range_notation += str(len(upload_data))
        logging_function('Uploading data now!')
        worksheet.spreadsheet.values_update = monkeypatch_values_update(worksheet.spreadsheet.values_update)
        worksheet.update(range_notation, upload_data)
        logging_function('Upload done.')



from airflow.models import BaseOperator
from airflow.hooks.base_hook import BaseHook
from airflow.utils.decorators import apply_defaults

from ewah.dwhooks import get_dwhook
from ewah.constants import EWAHConstants as EC
from ewah.ewah_utils.ssh_tunnel import start_ssh_tunnel

import hashlib

class EWAHBaseOperator(BaseOperator):
    """Extension of airflow's native Base Operator.

    EWAH operators always work in the same way: They extract raw data from a
    source and load it into a relational database aka data warehouse aka DWH.
    Transformations only happen insofar as they are required to bring data
    into a relational format.

    *How to use*
    The child class extracts data from the source. At the end of the child
    class' execute() function, it calls self.update(). Optional: The child can
    also repeatedly call the self.update() class for chunking.

    *Arguments of self.update():*
    Data is a list of dictionaries, where each list element will become
    one row of data and the contained dict is fieldname:value. Value can be
    None, or a fieldname can be missing in a particular dictionary, in which
    case the value is assumed to be None.

    columns_definition may be supplied at the self.update() function call,
    at initialization, or not at all, depending on the usecase. The use order is
    - If available, take the columns_definition from the self.update() call
    - If None, check if columns_definition was supplied to the operator at init
    - If neither, create a columns_definition on the fly using (all) the data

    columns_definition is a dictionary with fieldname:properties. Properties
    is None or a dictionary of option:value where option can be one of the
    following:
    - EC.QBC_FIELD_TYPE -> String: Field type (default text)
    - EC.QBC_FIELD_PK -> Boolean: Is this field the primary key? (default False)
    - EC.QBC_FIELD_NN -> Boolean: Not Null constraint for this field? (False)
    - EC.QBC_FIELD_UQ -> Boolean: Unique constraint for this field? (False)

    Note that the value of EC.QBC_FIELD_TYPE is DWH-engine specific!

    columns_definition is case sensitive.

    Implemented DWH engines:
    - PostgreSQL
    - Snowflake

    Next up:
    - BigQuery
    - Redshift
    """

    _IS_INCREMENTAL = False # Child class must update these values accordingly.
    _IS_FULL_REFRESH = False # Defines whether operator is usable in factories.

    _REQUIRES_COLUMNS_DEFINITION = False # raise error if true an none supplied

    _INDEX_QUERY = '''
        CREATE INDEX IF NOT EXISTS {0}
        ON "{1}"."{2}" ({3})
    '''

    upload_call_count = 0

    _metadata = {} # to be updated by operator, if applicable

    def __init__(
        self,
        source_conn_id,
        dwh_engine,
        dwh_conn_id,
        target_table_name,
        target_schema_name,
        target_schema_suffix='_next',
        target_database_name=None, # Only for Snowflake
        columns_definition=None,
        drop_and_replace=True,
        update_on_columns=None,
        primary_key_column_name=None,
        clean_data_before_upload=True,
        add_metadata=True, # adds columns with metadata to all rows
        # Note: that metadata is specified by a dict on operator level!
        # metadata can only be added if no columns definition is given
        exclude_columns=[], # list of columns to exclude, if no
        # columns_definition was supplied (e.g. for select * with sql)
        index_columns=[], # list of columns to create an index on. can be
        # an expression, must be quoted in list of quoting is required.
        source_ssh_tunnel_conn_id=None, # create SSH tunnel if set; uses host
        # and port from source_conn_id as remote host and port
        target_ssh_tunnel_conn_id=None, # see source_ssh_tunnel_conn_id
        hash_columns=None, # str or list of str - columns to hash pre-upload
        hashlib_func_name='sha256', # specify hashlib hashing function
    *args, **kwargs):
        super().__init__(*args, **kwargs)

        if hash_columns and not clean_data_before_upload:
            _msg = 'column hashing is only possible with data cleaning!'
            raise Exception(_msg)
        elif isinstance(hash_columns, str):
            hash_columns = [hash_columns]
        if hashlib_func_name:
            _msg = 'Invalid hashing function: hashlib.{0}()'
            _msg = _msg.format(hashlib_func_name)
            assert hasattr(hashlib, hashlib_func_name), _msg

        if columns_definition and exclude_columns:
            raise Exception('Must not supply both columns_definition and ' \
                + 'exclude_columns!')

        if not dwh_engine or not dwh_engine in EC.DWH_ENGINES:
            raise Exception('Invalid DWH Engine: {0}\n\nAccapted Engines:{1}'
                .format(
                str(dwh_engine),
                '\n'.join(EC.DWH_ENGINES),
            ))

        if index_columns and not dwh_engine == EC.DWH_ENGINE_POSTGRES:
            raise Exception('Indices are only allowed for PostgreSQL DWHs!')

        if dwh_engine == EC.DWH_ENGINE_SNOWFLAKE:
            if not target_database_name:
                conn_db_name = BaseHook.get_connection(dwh_conn_id)
                conn_db_name = conn_db_name.extra_dejson.get('database')
                if conn_db_name:
                    target_database_name = conn_db_name
                else:
                    raise Exception('If using DWH Engine {0}, must provide {1}!'
                        .format(
                            dwh_engine,
                            '"target_database_name" to specify the Database',
                        )
                    )
        else:
            if target_database_name:
                raise Exception('Received argument for "target_database_name"!')

        if self._REQUIRES_COLUMNS_DEFINITION:
            if not columns_definition:
                raise Exception('This operator requires the argument ' \
                    + 'columns_definition!')

        if primary_key_column_name and update_on_columns:
            raise Exception('Cannot supply BOTH primary_key_column_name AND' + \
                ' update_on_columns!')

        if not drop_and_replace:
            if not (
                update_on_columns
                or primary_key_column_name
                or (columns_definition and (0 < sum([
                        bool(columns_definition[col].get(
                            EC.QBC_FIELD_PK
                        )) for col in list(columns_definition.keys())
                    ])))
                ):
                raise Exception("If this is incremental loading of a table, "
                    + "one of the following is required:"
                    + "\n- List of columns to update on (update_on_columns)"
                    + "\n- Name of the primary key (primary_key_column_name)"
                    + "\n- Column definition (columns_definition) that includes"
                    + " the primary key(s)"
                )

        self.source_conn_id = source_conn_id
        self.dwh_engine = dwh_engine
        self.dwh_conn_id = dwh_conn_id
        self.target_table_name = target_table_name
        self.target_schema_name = target_schema_name
        self.target_schema_suffix = target_schema_suffix
        self.target_database_name = target_database_name
        self.columns_definition = columns_definition
        self.drop_and_replace = drop_and_replace
        if (not update_on_columns) and primary_key_column_name:
            if type(primary_key_column_name) == str:
                update_on_columns = [primary_key_column_name]
            elif type(primary_key_column_name) in (list, tuple):
                update_on_columns = primary_key_column_name
        self.update_on_columns = update_on_columns
        self.clean_data_before_upload = clean_data_before_upload
        self.primary_key_column_name = primary_key_column_name # may be used ...
        #   ... by a child class at execution!
        self.add_metadata = add_metadata
        self.exclude_columns = exclude_columns
        self.index_columns = index_columns
        self.source_ssh_tunnel_conn_id = source_ssh_tunnel_conn_id
        self.target_ssh_tunnel_conn_id = target_ssh_tunnel_conn_id
        self.hash_columns = hash_columns
        self.hashlib_func_name = hashlib_func_name

        self.hook = get_dwhook(self.dwh_engine)

    def execute(self, context):
        """ Why this method is defined here:
            When executing a task, airflow calls this method. Generally, this
            method contains the "business logic" of the individual operator.
            However, EWAH may want to do some actions for all operators. Thus,
            the child operators shall have an ewah_execute() function which is
            called by this general execute() method.
        """

        def close_ssh_tunnels():
            # close SSH tunnels if they exist
            if hasattr(self, 'source_ssh_tunnel_forwarder'):
                self.source_ssh_tunnel_forwarder.stop()
                del self.source_ssh_tunnel_forwarder
            if hasattr(self, 'target_ssh_tunnel_forwarder'):
                self.target_ssh_tunnel_forwarder.stop()
                del self.target_ssh_tunnel_forwarder

        # the upload hook is used in the self.upload_data() function
        # which is called by the child's ewah_execute function whenever there is
        # data to upload. If applicable: start SSH tunnel first!
        if self.target_ssh_tunnel_conn_id:
            self.log.info('Opening SSH tunnel to target...')
            self.target_ssh_tunnel_forwarder, self.dwh_conn = start_ssh_tunnel(
                ssh_conn_id=self.target_ssh_tunnel_conn_id,
                remote_conn_id=self.dwh_conn_id,
            )
        else:
            self.dwh_conn = BaseHook.get_connection(self.dwh_conn_id)
        self.upload_hook = self.hook(self.dwh_conn)

        # open SSH tunnel for the data source connection, if applicable
        if self.source_ssh_tunnel_conn_id:
            #if self.target_ssh_tunnel_conn_id:
            #    local_port = self.dwh_conn.port + 1
            #else:
            #    local_port = 0 # random assignment
            self.log.info('Opening SSH tunnel to source...')
            self.source_ssh_tunnel_forwarder, self.source_conn=start_ssh_tunnel(
                ssh_conn_id=self.source_ssh_tunnel_conn_id,
                remote_conn_id=self.source_conn_id,
            )
        elif self.source_conn_id:
            # resolve conn id here & delete the object to avoid usage elsewhere
            self.source_conn = BaseHook.get_connection(self.source_conn_id)
        del self.source_conn_id

        try:
            # execute operator
            result = self.ewah_execute(context)

            # if PostgreSQL and arg given: create indices
            for column in self.index_columns:
                # Use hashlib to create a unique 63 character string as index
                # name to avoid breaching index name length limits & accidental
                # duplicates / missing indices due to name truncation leading to
                # identical index names.
                self.hook.execute(self._INDEX_QUERY.format(
                    '__ewah_' + hashlib.blake2b(
                        (self.target_schema_name
                            + self.target_schema_suffix
                            + '.'
                            + self.target_table_name
                            + '.'
                            + column
                        ).encode(),
                        digest_size=28,
                    ).hexdigest(),
                    self.target_schema_name + self.target_schema_suffix,
                    self.target_table_name,
                    column,
                ))

            # commit only at the end, so that no data may be committed before an
            # error occurs.
            self.log.info('Now committing changes!')
            self.upload_hook.commit()
            self.upload_hook.close()
        except:
            # close SSH tunnels on failure before raising the error
            close_ssh_tunnels()
            raise

        # everything worked, now close SSH tunnels
        close_ssh_tunnels()

        return result

    def test_if_target_table_exists(self):
        hook = self.hook(self.dwh_conn)
        if self.dwh_engine == EC.DWH_ENGINE_POSTGRES:
            result = hook.test_if_table_exists(
                table_name=self.target_table_name,
                schema_name=self.target_schema_name + self.target_schema_suffix,
            )
            hook.close()
            return result
        elif self.dwh_engine == EC.DWH_ENGINE_SNOWFLAKE:
            result = hook.test_if_table_exists(
                table_name=self.target_table_name,
                schema_name=self.target_schema_name + self.target_schema_suffix,
                database_name=self.target_database_name,
            )
            hook.close()
            return result
        raise Exception('Function not implemented for DWH {0}!'.format(
            dwh_engine
        ))

    def _create_columns_definition(self, data):
        "Create a columns_definition from data (list of dicts)."
        inconsistent_data_type = EC.QBC_TYPE_MAPPING[self.dwh_engine].get(
            EC.QBC_TYPE_MAPPING_INCONSISTENT
        )
        def get_field_type(value):
            return EC.QBC_TYPE_MAPPING[self.dwh_engine].get(
                type(value)
            ) or inconsistent_data_type

        result = {}
        for datum in data:
            if self.add_metadata and self._metadata:
                datum.update(self._metadata)
            for field in datum.keys():
                if field in self.exclude_columns:
                    datum[field] = None
                elif field in (self.hash_columns or []) \
                    and not result.get(field):
                     # Type is appropriate string type & QBC_FIELD_HASH is true
                     result.update({field:{
                        EC.QBC_FIELD_TYPE: get_field_type('str'),
                        EC.QBC_FIELD_HASH: True,
                     }})
                elif not (result.get(field, {}).get(EC.QBC_FIELD_TYPE) \
                    == inconsistent_data_type) and (not datum[field] is None):
                    if result.get(field):
                        # column has been added in a previous iteration.
                        # If not default column: check if new and old column
                        #   type identification agree.
                        if not (result[field][EC.QBC_FIELD_TYPE] \
                            == get_field_type(datum[field])):
                            self.log.info(
                                'WARNING! Data types are inconsistent.'
                                + ' Affected column: {0}'.format(field)
                            )
                            result[field][EC.QBC_FIELD_TYPE] = \
                                inconsistent_data_type

                    else:
                        # First iteration with this column. Add to result.
                        result.update({field:{
                            EC.QBC_FIELD_TYPE: get_field_type(datum[field])
                        }})
        return result

    def upload_data(self, data=None, columns_definition=None):
        """Upload data, no matter the source. Call this functions in the child
            operator whenever data is available for upload, as often as needed.
        """
        if not data:
            self.log.info('No data to upload!')
            return
        self.upload_call_count += 1
        self.log.info('Chunk {1}: Uploading {0} rows of data.'.format(
            str(len(data)),
            str(self.upload_call_count),
        ))

        columns_definition = columns_definition or self.columns_definition
        if not columns_definition:
            self.log.info('Creating table schema on the fly based on data.')
            # Note: This is also where metadata is added, if applicable
            columns_definition = self._create_columns_definition(data)

        if self.update_on_columns:
            pk_list = self.update_on_columns # is a list already
        elif self.primary_key_column_name:
            pk_list = [self.primary_key_column_name]
        else:
            pk_list = []

        if pk_list:
            for pk_name in pk_list:
                if not pk_name in columns_definition.keys():
                    raise Exception(('Column {0} does not exist but is ' + \
                        'expected!').format(pk_name))
                columns_definition[pk_name][EC.QBC_FIELD_PK] = True

        hook = self.upload_hook

        if (not self.drop_and_replace) or (self.upload_call_count > 1):
            self.log.info('Checking for, and applying schema changes.')
            _new_schema_name = self.target_schema_name+self.target_schema_suffix
            new_cols, del_cols = hook.detect_and_apply_schema_changes(
                new_schema_name=_new_schema_name,
                new_table_name=self.target_table_name,
                new_columns_dictionary=columns_definition,
                # When introducing a feature utilizing this, remember to
                #  consider multiple runs within the same execution
                drop_missing_columns=False and self.upload_call_count==1,
                database=self.target_database_name,
                commit=False, # Commit only when / after uploading data
            )
            self.log.info('Added fields:\n\t{0}\nDeleted fields:\n\t{1}'.format(
                '\n\t'.join(new_cols) or '\n',
                '\n\t'.join(del_cols) or '\n',
            ))

        self.log.info('Uploading data now.')
        hook.create_or_update_table(
            data=data,
            columns_definition=columns_definition,
            table_name=self.target_table_name,
            schema_name=self.target_schema_name,
            schema_suffix=self.target_schema_suffix,
            database_name=self.target_database_name,
            drop_and_replace=self.drop_and_replace and \
                (self.upload_call_count == 1), # In case of chunking of uploads
            update_on_columns=self.update_on_columns,
            commit=False, # See note below for reason
            logging_function=self.log.info,
            clean_data_before_upload=self.clean_data_before_upload,
            hash_columns=self.hash_columns,
            hashlib_func_name=self.hashlib_func_name,
        )
        """ Note on committing changes:
            The hook used for data uploading is created at the beginning of the
            execute function and automatically committed and closed at the end.
            DO NOT commit in this function, as multiple uploads may be required,
            and any intermediate commit may be subsequently followed by an
            error, which would then result in incomplete data committed.
        """

class EWAHEmptyOperator(EWAHBaseOperator):
    _IS_INCREMENTAL = True
    _IS_FULL_REFRESH = True
    def __init__(self, *args, **kwargs):
        raise Exception('Failed to load operator! Probably missing' \
            + ' requirements for the operator in question.\n\nSupplied args:' \
            + '\n\t' + '\n\t'.join(args) + '\n\nSupplied kwargs:\n\t' \
            + '\n\t'.join(['{0}: {1}'.format(k, v) for k, v in kwargs.items()])
        )



from ewah.operators.base_operator import EWAHBaseOperator
from ewah.ewah_utils.airflow_utils import airflow_datetime_adjustments
from ewah.constants import EWAHConstants as EC

from airflow.hooks.base_hook import BaseHook

from datetime import datetime, timedelta

from requests.auth import HTTPBasicAuth
import requests
import json
import time

class EWAHZendeskOperator(EWAHBaseOperator):

    _IS_INCREMENTAL = True
    _IS_FULL_REFRESH = False

    _base_url = 'https://{support_url}.zendesk.com/{endpoint}'
    # _base_url = \
        # 'https://{support_url}.zendesk.com/api/v2/incremental/{resource}.json'

    def __init__(self,
        support_url,
        resource,
        auth_type,
        data_from=None,
    *args, **kwargs):

        self._accepted_resources = {
            'tickets': {
                'function': self.execute_incremental_cursor_based,
                'datetime_field': 'updated_at',
            },
            'ticket_audits': {
                'function': self.execute_incremental_cursor_based,
                'datetime_field': 'created_at',
            },
            'ticket_metric_events': {
                'function': self.execute_incremental_time_based,
            },
            'users': {
                'function': self.execute_incremental_time_based,
            },
            'ticket_fields': {
                'function': self.get_custom_ticket_fields,
                'drop_and_replace': True,
            },
        }

        if not resource in self._accepted_resources.keys():
            raise Exception('resource "{0}" not supported!'.format(resource))

        if not auth_type in ['basic_auth']:
            raise Exception('auth_type must be basic_auth!')

        #if not type(page_limit) == int or page_limit < 1 or page_limit > 100:
        #    raise Exception('page_limit must be a positive interger <= 100!')

        if not type(data_from) == datetime:
            raise Exception('data_from must be a datetime object!')

        kwargs['primary_key_column_name'] = \
            kwargs.get('primary_key_column_name', 'id')

        if self._accepted_resources[resource].get('drop_and_replace'):
            kwargs['drop_and_replace'] = True

        super().__init__(*args, **kwargs)

        self.support_url = support_url
        self.resource = resource
        self.auth_type = auth_type
        self.data_from = data_from
        #self.page_limit = page_limit

    def ewah_execute(self, context):
        # Make sure it is at least 70 seconds after next_execution_date!
        # Reason: Immediate execution may miss the latest tickets
        td70 = timedelta(seconds=70)
        while datetime.now() < (context['next_execution_date'] + td70):
            self.log.info('delaying execution...')
            time.sleep(5)

        self.data_from = self.make_unix_datetime(self.get_data_from(context))
        self.data_until= self.make_unix_datetime(context['next_execution_date'])

        conn = self.source_conn
        if self.auth_type == 'basic_auth':
            self.auth = HTTPBasicAuth(conn.login, conn.password)
        else:
            raise Exception('auth_type not implemented!')

        self._metadata.update({'support_url': self.support_url})

        # run correct execute function
        return self._accepted_resources[self.resource]['function'](context,self)

    def get_data_from(self, context):
        if self.test_if_target_table_exists():
            return context['execution_date']
        else:
            return self.data_from or context['execution_date']

    def make_unix_datetime(self, dt):
        return str(int(time.mktime(dt.timetuple())))

    def get_custom_ticket_fields(_, context, self):
        url = self._base_url.format(
            support_url=self.support_url,
            endpoint='api/v2/ticket_fields.json',
        )
        req = requests.get(url, auth=self.auth)
        if not req.status_code == 200:
            raise Exception('Error {1} when calling Zendesk API: {0}'.format(
                req.text,
                str(req.status_code),
            ))
        data = json.loads(req.text)['ticket_fields']
        self.upload_data(data)

    def execute_incremental_cursor_based(_, context, self):
        if not self.resource in ['tickets', 'ticket_audits']:
            raise Exception('cursor-based load not implemented!')

        if self.resource == 'tickets':
            endpoint = 'api/v2/incremental/tickets/cursor.json'
            params = {
                'start_time': self.data_from,
                'include': 'metric_sets',
            }
            response_resource = 'tickets'
        elif self.resource == 'ticket_audits':
            endpoint = 'api/v2/ticket_audits.json'
            params = {
                'start_time': self.data_from,
            }
            response_resource = 'audits'

        datetime_field = self._accepted_resources[self.resource]
        datetime_field = datetime_field['datetime_field']

        url = self._base_url.format(
            support_url=self.support_url,
            endpoint=endpoint,
        )

        first_call = True
        while first_call or req.status_code == 200:
            if first_call:
                first_call = False
            self.log.info('Requesting a page of data...')
            req = requests.get(url, params=params, auth=self.auth)
            response = json.loads(req.text)
            if req.status_code == 200:
                data = response[response_resource]
                if data:
                    last_updated_at = datetime.strptime(
                        data[-1][datetime_field],
                        '%Y-%m-%dT%H:%M:%SZ',
                    )
                    self.upload_data(data)
                if response.get('end_of_stream') or \
                    last_updated_at > context['next_execution_date']:
                    break
                params = {'cursor': response.get('after_cursor')}

        if not req.status_code == 200:
            raise Exception('Error {1} when calling Zendesk API: {0}'.format(
                req.text,
                str(req.status_code),
            ))


    def execute_incremental_time_based(_, context, self):

        url = self._base_url.format(**{
            'support_url': self.support_url,
            'endpoint': 'api/v2/incremental/{0}.json'.format(self.resource)
        })

        params = {
            'start_time': self.data_from,
        }

        # Time-based incremental exports, run the first, the paginate!
        self.log.info('Calling Zendesk api.\nurl: {0}\nparams:{1}'.format(
            url,
            str(params),
        ))
        r = requests.get(url, params=params, auth=self.auth)
        data = json.loads(r.text)
        while r.status_code == 200 \
            and not data.get('end_of_stream') \
            and data.get('end_time') \
            and data.get('end_time') <= int(self.data_until):
            self.upload_data(data[self.resource]) # uploads previous request
            self.log.info('Requesting next page of data...')
            r = requests.get(data['next_page'], auth=self.auth) # new request
            data = json.loads(r.text)

        if not r.status_code == 200:
            raise Exception('Error {1} when calling Zendesk API: {0}'.format(
                r.text,
                str(r.status_code),
            ))

        # upload final request
        self.upload_data(data[self.resource])



from ewah.operators.base_operator import EWAHBaseOperator
from ewah.ewah_utils.airflow_utils import airflow_datetime_adjustments
from ewah.constants import EWAHConstants as EC

from airflow.hooks.S3_hook import S3Hook

from datetime import datetime, timedelta

import json
import csv

class EWAHS3Operator(EWAHBaseOperator):
    """Only implemented for JSON and CSV files from S3 right now!"""

    template_fields = ('data_from', 'data_until')

    _IS_INCREMENTAL = True
    _IS_FULL_REFRESH = True

    _IMPLEMENTED_FORMATS = [
        'JSON',
        'AWS_FIREHOSE_JSON',
        'CSV',
    ]

    _BOM = b'\xef\xbb\xbf'

    def __init__(self,
        bucket_name,
        file_format,
        prefix='', # use for subfolder structures
        suffix=None, # use e.g. for file types
        key_name=None, # if not specified, uses data_from and data_until
        data_from=None,
        data_until=None,
        csv_format_options={},
        csv_encoding='utf-8',
    *args, **kwargs):

        if not file_format in self._IMPLEMENTED_FORMATS:
            raise Exception('File format not implemented! Available formats:' \
                + ' {0}'.format(str(self._IMPLEMENTED_FORMATS)))

        if key_name:
            if data_from or data_until:
                raise Exception('Both key_name and either data_from or data_' \
                    + 'until specified! Cannot have both specified.')

        if not file_format == 'CSV' and csv_format_options:
            raise Exception('csv_format_options is only valid for CSV files!')



        #if not self.drop_and_replace:
        #    if not (data_from and data_until):
        #        raise Exception('For incremental loading, you must specify ' \
        #            + 'both data_from and data_until')
        # if incrementally loading with data_from and data_until: default
        #   to execution dates!

        self.bucket_name = bucket_name
        self.prefix = prefix
        self.suffix = suffix
        self.key_name = key_name
        self.data_from = data_from
        self.data_until = data_until
        self.file_format = file_format
        self.csv_format_options = csv_format_options
        self.csv_encoding = csv_encoding

        super().__init__(*args, **kwargs)

    def _iterate_through_bucket(self, s3hook, bucket, prefix):
        """ The bucket.objects.filter() method only returns a max of 1000
            objects. If more objects are in an S3 bucket, pagniation is
            required. See also: https://stackoverflow.com/questions/44238525/how-to-iterate-over-files-in-an-s3-bucket
        """
        cli = s3hook.get_client_type('s3')
        paginator = cli.get_paginator('list_objects_v2')
        page_iterator = paginator.paginate(Bucket=bucket, Prefix=prefix)

        for page in page_iterator:
            if page['KeyCount'] > 0:
                for item in page['Contents']:
                    yield item

    def ewah_execute(self, context):
        if not self.drop_and_replace and not self.key_name:
            self.data_from = self.data_from or context['execution_date']
            self.data_until = self.data_until or context['next_execution_date']

        if self.file_format == 'JSON':
            return self.execute_json(
                context=context,
                f_get_data=lambda hook, key, bucket: json.loads(
                    hook.read_key(key, bucket)
                ),
            )
        elif self.file_format == 'AWS_FIREHOSE_JSON':
            return self.execute_json(
                context=context,
                f_get_data=lambda hook, key, bucket: json.loads(
                    '['+hook.read_key(key, bucket).replace('}{', '},{')+']'
                ),
            )
        elif self.file_format == 'CSV':
            return self.execute_csv(
                context=context,
            )
        else:
            raise Exception('File format not implemented!')

    def execute_csv(self, context):
        self.data_from = airflow_datetime_adjustments(self.data_from)
        self.data_until = airflow_datetime_adjustments(self.data_until)
        hook = S3Hook(self.source_conn.conn_id)
        suffix = self.suffix
        if self.key_name:
            data = hook.read_key(self.key_name, self.bucket_name)
            self.upload_data(data=data)
        else:
            objects = self._iterate_through_bucket(
                s3hook=hook,
                bucket=self.bucket_name,
                prefix=self.prefix,
            )
            for obj_iter in objects:
                # skip all files outside of loading scope
                obj = hook.get_key(obj_iter['Key'], self.bucket_name)
                if self.data_from and obj.last_modified < self.data_from:
                    continue
                if self.data_until and obj.last_modified >= self.data_until:
                    continue
                if suffix and not suffix == obj.key[-len(suffix):]:
                    continue

                self.log.info('Loading data from file {0}'.format(obj.key))
                raw_data = obj.get()['Body'].read()
                # remove BOM if it exists
                # also, if file has a BOM, it is 99.9% utf-9 encoded!
                if raw_data[:3] == self._BOM:
                    raw_data = raw_data[3:]
                    csv_encoding = 'utf-8'
                else:
                    csv_encoding = self.csv_encoding
                # there may be a BOM while still not utf-8 -> use the
                # given csv_encoding argument if so
                try:
                    raw_data = raw_data.decode(csv_encoding).splitlines()
                except UnicodeDecodeError:
                    if csv_encoding == self.csv_encoding:
                        raise
                    raw_data = raw_data.decode(self.csv_encoding).splitlines()
                reader = csv.DictReader(raw_data, **self.csv_format_options)
                data = list(reader)
                self._metadata.update({
                    'bucket_name': self.bucket_name,
                    'file_name': obj.key,
                    'file_last_modified': str(obj.last_modified),
                })
                self.upload_data(data=data)

    def execute_json(self, context, f_get_data):
        self.data_from = airflow_datetime_adjustments(self.data_from)
        self.data_until = airflow_datetime_adjustments(self.data_until)

        hook = S3Hook(self.source_conn.conn_id)
        suffix = self.suffix

        if self.key_name:
            data = f_get_data(
                hook=hook,
                key=self.key_name,
                bucket=self.bucket_name,
            )
            self.upload_data(data=data)
        else:
            data = []
            objects = self._iterate_through_bucket(
                s3hook=hook,
                bucket=self.bucket_name,
                prefix=self.prefix,
            )
            for obj_iter in objects:
                obj = hook.get_key(obj_iter['Key'], self.bucket_name)
                if self.data_from and obj.last_modified < self.data_from:
                    continue
                if self.data_until and obj.last_modified >= self.data_until:
                    continue
                if suffix and not suffix == obj.key[-len(suffix):]:
                    continue

                self.log.info('Loading data from file {0}'.format(
                    obj.key,
                ))
                self._metadata.update({
                    'bucket_name': self.bucket_name,
                    'file_name': obj.key,
                    'file_last_modified': str(obj.last_modified),
                })
                data = f_get_data(
                    hook=hook,
                    key=obj.key,
                    bucket=self.bucket_name,
                )
                self.upload_data(data=data)



from ewah.dwhooks.base_dwhook import EWAHBaseDWHook
from ewah.constants import EWAHConstants as EC

from psycopg2.extras import execute_values
from psycopg2 import connect as pg_connect

class EWAHDWHookPostgres(EWAHBaseDWHook):

    _QUERY_SCHEMA_CHANGES_COLUMNS = """
        SELECT
        	f.attname AS "name"
        FROM pg_attribute f
        	JOIN pg_class cl ON cl.OID = f.attrelid
        	LEFT JOIN pg_namespace n ON n.OID = cl.relnamespace
        WHERE cl.relkind = 'r'::CHAR
        	AND n.nspname = %(schema_name)s
        	AND cl.relname = %(table_name)s
        	AND f.attnum > 0;
    """
    _QUERY_SCHEMA_CHANGES_DROP_COLUMN = """
        ALTER TABLE "{schema_name}"."{table_name}"
        DROP COLUMN IF EXISTS "{column_name}" CASCADE;
    """
    _QUERY_SCHEMA_CHANGES_ADD_COLUMN = """
        ALTER TABLE "{schema_name}"."{table_name}"
        ADD COLUMN "{column_name}" {column_type};
    """
    _QUERY_TABLE = 'SELECT * FROM "{schema_name}"."{table_name}"'

    def __init__(self, *args, **kwargs):
        super().__init__(EC.DWH_ENGINE_POSTGRES, *args, **kwargs)

    def _create_conn(self):
        return pg_connect(
            "dbname='{0}' user='{1}' host='{2}' password='{3}' port='{4}'"
            .format(
                self.credentials.schema,
                self.credentials.login,
                self.credentials.host,
                self.credentials.password,
                self.credentials.port,
            )
        )

    def _create_or_update_table(
        self,
        data,
        table_name,
        schema_name,
        schema_suffix,
        # database_name,
        columns_definition,
        columns_partial_query,
        update_on_columns,
        drop_and_replace,
        logging_function,
        upload_chunking=100000,
        pk_columns=[],
    ):
        logging_function('Preparing DWH Tables...')
        schema_name += schema_suffix
        if drop_and_replace or (not self.test_if_table_exists(
            table_name=table_name,
            schema_name=schema_name,
        )):
            self.execute(
                sql="""
                    DROP TABLE IF EXISTS "{schema_name}"."{table_name}" CASCADE;
                    CREATE TABLE "{schema_name}"."{table_name}" ({columns});
                """.format(**{
                    'schema_name': schema_name,
                    'table_name': table_name,
                    'columns': columns_partial_query,
                }),
                commit=False,
            )
            if pk_columns:
                self.execute(
                    sql="""
                        ALTER TABLE ONLY "{schema_name}"."{table_name}"
                        ADD PRIMARY KEY ("{columns}");
                    """.format(
                        schema_name=schema_name,
                        table_name=table_name,
                        columns='","'.join(pk_columns),
                    )
                )

        if not drop_and_replace and update_on_columns:
            # make sure there is a unique constraint for update_on_columns
            self.execute(
                sql="""
                    ALTER TABLE "{schema_name}"."{table_name}"
                    DROP CONSTRAINT IF EXISTS "{constraint}";
                    ALTER TABLE "{schema_name}"."{table_name}"
                    ADD CONSTRAINT "{constraint}" UNIQUE ("{columns}");
                """.format(**{
                    'schema_name': schema_name,
                    'table_name': table_name,
                    'constraint': 'ufu_{0}_{1}'.format(schema_name, table_name),
                    'columns': '", "'.join(update_on_columns),
                }),
                commit=False,
            )

        set_columns = []
        for column in columns_definition.keys():
            if not (column in update_on_columns):
                set_columns += [column]

        cols_list = list(columns_definition.keys())
        sql="""
            INSERT INTO "{schema_name}"."{table_name}"
            ("{column_names}") VALUES {placeholder}
            ON CONFLICT {do_on_conflict};
        """.format(**{
            'schema_name': schema_name,
            'table_name': table_name,
            'placeholder': '{placeholder}',
            'column_names': '", "'.join(cols_list),
            'do_on_conflict': 'DO NOTHING' \
                if drop_and_replace or not update_on_columns else """
                ("{update_on_columns}") DO UPDATE SET\n\t{sets}
            """.format(
                update_on_columns='", "'.join(update_on_columns),
                sets='\n\t,'.join([
                    '"{column}" = EXCLUDED."{column}"'.format(column=column)
                    for column in set_columns
                ]),
            ),
        }).replace('%', '%%').format(**{
            'placeholder': '%s',
        })
        logging_function('Now Uploading! Using SQL:\n\n{0}'.format(sql))
        # escape crappy column names by using aliases in the psycopg2 template
        cols_map = {cols_list[i]:'col_'+str(i) for i in range(len(cols_list))}
        template = '(%('
        template += ')s, %('.join([value for key, value in cols_map.items()])
        template += ')s)'
        cur = self.get_cursor()
        while data:
            temp_data = data[:upload_chunking]
            data = data[upload_chunking:]
            upload_data = []
            while temp_data:
                upload_data += [{
                    cols_map[key]:val for key, val in temp_data.pop(0).items()
                }]
            execute_values(
                cur=cur,
                sql=sql,
                argslist=upload_data,
                template=template,
            )
        logging_function('Upload done.')

        self.execute(
            sql='ANALYZE "{0}"."{1}";'.format(schema_name, table_name),
            commit=False,
        )


    def commit(self):
        self.logging_func('DWH Hook: Committing changes!')
        self.conn.commit()

    def rollback(self):
        self.logging_func('DWH Hook: Rolling back changes!')
        self.conn.rollback()

    def test_if_table_exists(self, table_name, schema_name):
        return bool(self.execute_and_return_result(
                sql="SELECT to_regclass(%(name)s);",
                params={'name':'"{0}"."{1}"'.format(schema_name, table_name)},
            )[0][0])



from ewah.operators.sql_base_operator import EWAHSQLBaseOperator
from ewah.constants import EWAHConstants as EC

from airflow.hooks.base_hook import BaseHook

import cx_Oracle

class EWAHOracleSQLOperator(EWAHSQLBaseOperator):

        _SQL_BASE = 'SELECT\n{columns}\nFROM "{schema}"."{table}"\nWHERE {where_clause}'
        _SQL_BASE_SELECT = \
            'SELECT * FROM ({select_sql}) t WHERE {{0}}'
        _SQL_COLUMN_QUOTE = '"'
        _SQL_MINMAX_CHUNKS = '''
            WITH base AS ({base})
            SELECT MIN({column}), MAX({column})
            FROM base;
        '''
        _SQL_CHUNKING_CLAUSE = '''
            AND {column} >= :from_value
            AND {column} <{equal_sign} :until_value
        '''
        _SQL_PARAMS = ':{0}'

        def __init__(self, *args, **kwargs):
            self.sql_engine = self._ORACLE
            super().__init__(*args, **kwargs)

        def _get_data_from_sql(self,
                sql,
                params={},
                return_dict=True,
            ):
            '''In Oracle, params are passed to the execute() function as kwargs
            https://cx-oracle.readthedocs.io/en/latest/user_guide/bind.html
            Params are then referenced like so:
                SELECT * FROM table WHERE field1 = :val1 AND field2 = :val2
            with a params dict structured like so: {'val1': 1, 'val2': 3}
            Also see here: https://stackoverflow.com/questions/35045879/cx-oracle-how-can-i-receive-each-row-as-a-dictionary
            to understand the workaround regarding return_dict
            '''
            def makeDictFactory(cursor):
                columnNames = [d[0] for d in cursor.description]
                def createRow(*args):
                    return dict(zip(columnNames, args))
                return createRow

            oracle_conn = cx_Oracle.connect(
                self.source_conn.login,
                self.source_conn.password,
                '{0}:{1}/{2}'.format(
                    self.source_conn.host,
                    self.source_conn.port,
                    self.source_conn.schema,
                ),
                encoding='UTF-8'
            )
            cursor = oracle_conn.cursor()
            if sql.strip()[-1:] == ';': # OracleSQL doesn't like semicolons
                sql = sql.strip()[:-1]
            self.log.info('Executing:\n{0}\n\nWith params:\n{1}'.format(
                sql,
                str(params),
            ))
            cursor.execute(sql, **params)
            if return_dict:
                cursor.rowfactory = makeDictFactory(cursor)
            data = cursor.fetchall()
            cursor.close()
            oracle_conn.close()
            return data



from ewah.operators.base_operator import EWAHBaseOperator
from ewah.ewah_utils.airflow_utils import airflow_datetime_adjustments
from ewah.constants import EWAHConstants as EC

from airflow.hooks.base_hook import BaseHook

import googlemaps

class EWAHGMapsOperator(EWAHBaseOperator):
    """ Gets address data from a custom SQL fired against the DWH.

    Params:
        - address_sql - a (templated) SQL statement that fetches the data
            as required; only one column called 'address' will be respected.
            Any additional columns will be ignored, and failure to provide
            such a column will cause failure of the operator.
    """

    template_fields = ('address_sql',)

    _IS_INCREMENTAL = True
    _IS_FULL_REFRESH = True

    def __init__(self,
        address_sql,
    *args, **kwargs):

        kwargs['primary_key_column_name'] = 'address'
        super().__init__(*args, **kwargs)

        self.address_sql = address_sql

    def ewah_execute(self, context):
        dwh_hook = self.upload_hook

        addresses = dwh_hook.execute_and_return_result(
            sql=self.address_sql,
            return_dict=True,
        )

        api_key = self.source_conn.password

        client = googlemaps.Client(key=api_key)

        data = []
        while addresses:
            try:
                address = addresses.pop(0)['address']
            except:
                raise Exception('Your SQL did not provide an address column!')

            data += [{
                'address': address,
                'geocode_result': client.geocode(address),
                }]

        self.upload_data(data)



from datetime import datetime, date

class EWAHConstants:
    "This class contains a number of constants for use throughout Ewah."

    # Available DWH Engines
    DWH_ENGINE_POSTGRES = 'PostgreSQL'
    DWH_ENGINE_SNOWFLAKE = 'Snowflake'
    DWH_ENGINE_BIGQUERY = 'BigQuery'
    DWH_ENGINE_REDSHIFT = 'Redshift'
    DWH_ENGINE_S3 = 'S3'
    DWH_ENGINE_GS = 'GoogleSheets'
    DWH_ENGINES = [
        DWH_ENGINE_POSTGRES,
        DWH_ENGINE_SNOWFLAKE,
        # DWH_ENGINE_BIGQUERY, To Be Implemented
        # DWH_ENGINE_REDSHIFT, To Be Implemented
        # DWH_ENGINE_S3, To Be Implemented
        DWH_ENGINE_GS,
    ]

    # Query Building Constants
    QBC_FIELD_TYPE = 'data_type'

    QBC_FIELD_PK = 'is_primary_key'
    QBC_FIELD_NN = 'is_not_null'
    QBC_FIELD_UQ = 'is_unique'
    QBC_FIELD_HASH = 'is_hash_column'
    QBC_FIELD_GSHEET_COLNO = 'column' # Gsheet operator: position of the column
    QBC_FIELD_CONSTRAINTS_MAPPING = {
        DWH_ENGINE_POSTGRES: {
            QBC_FIELD_PK: 'PRIMARY KEY',
            QBC_FIELD_NN: 'NOT NULL',
            QBC_FIELD_UQ: 'UNIQUE',
        },
        DWH_ENGINE_SNOWFLAKE: {
            QBC_FIELD_PK: 'PRIMARY KEY',
            QBC_FIELD_NN: 'NOT NULL',
            QBC_FIELD_UQ: 'UNIQUE',
        },
        # DWH_ENGINE_BIGQUERY: {},
        # DWH_ENGINE_REDSHIFT: {},
        # DWH_ENGINE_S3: {},
        DWH_ENGINE_GS: {
            # Not applicable
        },
    }

    """When a columns_definition is given, but a column does not have a field
    type defined, use default as specified here. If no columns_definition is
    given, and a field is of inconsistent data type or type does not exist
    in the mapping below, use inconsistent value specified here."""
    QBC_TYPE_MAPPING_DEFAULT = 'tm_default'
    QBC_TYPE_MAPPING_INCONSISTENT = 'tm_incon'
    QBC_TYPE_MAPPING = {
        DWH_ENGINE_POSTGRES: {
            QBC_TYPE_MAPPING_DEFAULT: 'text',
            QBC_TYPE_MAPPING_INCONSISTENT: 'text',
            str: 'text',
            int: 'bigint',
            float: 'numeric',
            dict: 'jsonb',
            list: 'jsonb',
            tuple: 'jsonb',
            set: 'jsonb',
            frozenset: 'jsonb',
            bool: 'boolean',
            datetime: 'timestamp with time zone',
            date: 'date',
        },
        DWH_ENGINE_SNOWFLAKE: {
            QBC_TYPE_MAPPING_DEFAULT: 'text',
            QBC_TYPE_MAPPING_INCONSISTENT: 'text',
            str: 'text',
            int: 'bigint',
            float: 'numeric',
            dict: 'jsonb',
            list: 'jsonb',
            tuple: 'jsonb',
            set: 'jsonb',
            frozenset: 'jsonb',
            bool: 'boolean',
            datetime: 'timestamp with time zone',
            date: 'date',
        },
        DWH_ENGINE_BIGQUERY: {
            QBC_TYPE_MAPPING_DEFAULT: 'STRING',
            QBC_TYPE_MAPPING_INCONSISTENT: 'STRING',
            str: 'STRING',
            int: 'INT64',
            float: 'NUMERIC',
            dict: 'STRUCT',
            list: 'STRUCT',
            tuple: 'STRUCT',
            set: 'STRUCT',
            frozenset: 'STRUCT',
            bool: 'BOOL',
            datetime: 'TIMESTAMP',
            date: 'DATE',
            bytes: 'BYTES',
        },
        # DWH_ENGINE_REDSHIFT: {},
        # DWH_ENGINE_S3: {},
        DWH_ENGINE_GS: { # must have these two values evaluate to True as bool
            QBC_TYPE_MAPPING_DEFAULT: 'x',
            QBC_TYPE_MAPPING_INCONSISTENT: 'x',
        },
    }



from ewah.dwhooks.base_dwhook import EWAHBaseDWHook
from ewah.constants import EWAHConstants as EC

import os
import csv
import snowflake.connector
from tempfile import NamedTemporaryFile
from airflow.utils.file import TemporaryDirectory
from airflow.models import BaseOperator

class SnowflakeOperator(BaseOperator):
    "Operate to execute SQL on Snowflake"

    def __init__(
        self,
        sql,
        snowflake_conn_id,
        database,
        params=None,
    *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.sql = sql
        self.conn_id = snowflake_conn_id
        self.database = database
        # an attribute called "params" of an operator leads to an airflow error
        # if that attribute is None due to default templating of "task.params"!
        self._params = params

    def execute(self, context):
        hook = EWAHDWHookSnowflake(
            dwh_conn=EWAHBaseDWHook.get_connection(self.conn_id),
            database=self.database,
        )
        self.log.info('execute: {0}'.format(self.sql))
        hook.execute(self.sql, commit=True, params=self._params)
        hook.close()

class EWAHDWHookSnowflake(EWAHBaseDWHook):

    _QUERY_SCHEMA_CHANGES_COLUMNS = """
        SELECT
            column_name
        FROM "{database_name}".information_schema.columns
        WHERE table_catalog = %(database_name)s
            AND table_schema = %(schema_name)s
            AND table_name = %(table_name)s
        ORDER BY ordinal_position ASC;
    """
    _QUERY_SCHEMA_CHANGES_DROP_COLUMN = """
        ALTER TABLE "{database_name}"."{schema_name}"."{table_name}"
        DROP COLUMN IF EXISTS "{column_name}" CASCADE;
    """
    _QUERY_SCHEMA_CHANGES_ADD_COLUMN = """
        ALTER TABLE "{database_name}"."{schema_name}"."{table_name}"
        ADD COLUMN "{column_name}" {column_type};
    """
    _QUERY_TABLE = """
        SELECT * FROM "{database_name}"."{schema_name}"."{table_name}"
    """

    def __init__(self, *args, database=None, **kwargs):
        self.database = database
        super().__init__(EC.DWH_ENGINE_SNOWFLAKE, *args, **kwargs)


    def _create_conn(self, database=None):
        creds = self.credentials
        extra = creds.extra_dejson
        extra['password'] = creds.password
        if creds.host:
            extra['account'] = creds.host
        if creds.login:
            extra['user'] = creds.login
        if database or self.database:
            extra['database'] = database or self.database

        return snowflake.connector.connect(**extra)

    def _create_or_update_table(
        self,
        data,
        table_name,
        schema_name,
        schema_suffix,
        database_name,
        columns_definition,
        columns_partial_query,
        update_on_columns,
        drop_and_replace,
        logging_function,
        pk_columns=[],
    ):
        logging_function('Preparing DWH Tables...')
        schema_name += schema_suffix
        new_table_name = table_name + '_new'
        self.execute(
            sql="""CREATE OR REPLACE TABLE
                    "{database_name}"."{schema_name}"."{table_name}"
                    ({columns});
            """.format(**{
                'database_name': database_name,
                'schema_name': schema_name,
                'table_name': new_table_name,
                'columns': columns_partial_query,
            }),
            commit=False,
        )

        list_of_columns = self.execute_and_return_result(
            sql="""SELECT
                column_name
                FROM "{0}".information_schema.columns
                WHERE table_catalog = '{0}'
                AND table_schema = '{1}'
                AND table_name = '{2}'
                ORDER BY ordinal_position ASC
            """.format(
                database_name,
                schema_name,
                new_table_name,
            ),
        )
        list_of_columns = [col[0].strip() for col in list_of_columns]

        self.log.info('Writing data to a temporary .csv file')
        with TemporaryDirectory(prefix='uploadtosnowflake') as tmp_dir:
            with NamedTemporaryFile(
                dir=tmp_dir,
                prefix=new_table_name,
                suffix='.csv',
            ) as datafile:
                file_name = os.path.abspath(datafile.name)
                with open(file_name, mode='w') as csv_file:
                    csvwriter = csv.writer(
                        csv_file,
                        delimiter=',',
                        quotechar='"',
                        quoting=csv.QUOTE_MINIMAL,
                    )

                    for _ in range(len(data)):
                        datum = data.pop(0)
                        # Make sure order of csv is the same as order of columns
                        csvwriter.writerow(
                            [datum.get(col) for col in list_of_columns],
                        )

                # now stage and copy into snowflake!
                sql_upload = '''
                    USE DATABASE "{2}";
                    USE SCHEMA "{3}";
                    DROP FILE FORMAT IF EXISTS {1}_format;
                    CREATE OR REPLACE FILE FORMAT {1}_format
                        type = 'CSV'
                        field_delimiter = ','
                        field_optionally_enclosed_by = '"'
                    ;
                    DROP STAGE IF EXISTS {1}_stage;
                    CREATE OR REPLACE STAGE {1}_stage
                        file_format = {1}_format;
                    PUT file://{0} @{1}_stage;
                    COPY INTO "{2}"."{3}"."{1}" FROM '@{1}_stage/{4}.gz';
                '''.format(
                    file_name,
                    new_table_name,
                    database_name,
                    schema_name,
                    file_name[file_name.rfind('/')+1:],
                )
                self.log.info('Uploading data to Snowflake...')
                self.execute(sql_upload)

        if drop_and_replace or (not self.test_if_table_exists(
            table_name=table_name,
            schema_name=schema_name,
            database_name=database_name,
        )):
            sql_final = '''
                USE SCHEMA "{0}"."{1}";
                DROP TABLE IF EXISTS "{0}"."{1}"."{2}" CASCADE;
                ALTER TABLE "{0}"."{1}"."{3}" RENAME TO "{2}";
            '''.format(
                database_name,
                schema_name,
                table_name,
                new_table_name,
            )
            if pk_columns:
                sql_final += '''
                    ALTER TABLE "{0}"."{1}"."{2}"
                    ADD PRIMARY KEY ("{3}");
                '''.format(
                    database_name,
                    schema_name,
                    table_name,
                    '","'.join(pk_columns),
                )
        else:
            update_set_cols = []
            for col in columns_definition.keys():
                if not (col in update_on_columns):
                    update_set_cols += [col]

            sql_final = '''
                USE SCHEMA "{0}"."{1}";
                MERGE INTO "{0}"."{1}"."{2}" AS a
                    USING "{0}"."{1}"."{3}" AS b
                    ON {4}
                    WHEN MATCHED THEN UPDATE
                    SET {5}
                    WHEN NOT MATCHED THEN INSERT
                    ({6}) VALUES ({7})
                    ;
                DROP TABLE "{0}"."{1}"."{3}" CASCADE;
            '''.format(
                database_name,
                schema_name,
                table_name,
                new_table_name,
                ' AND '.join([
                    'a."{0}" = b."{0}"'.format(col) for col in update_on_columns
                ]),
                ', '.join([
                    'a."{0}" = b."{0}"'.format(col) for col in update_set_cols
                ]),
                '"'+'", "'.join(list(columns_definition.keys()))+'"',
                ', '.join([
                    'b."{0}"'.format(col) \
                    for col in list(columns_definition.keys())
                ]),
            )

        self.log.info('Final Step: Merging data')
        self.execute(sql_final)

    def commit(self):
        self.cur.execute('COMMIT;')
        self.cur.execute('BEGIN;')

    def rollback(self):
        self.cur.execute('ROLLBACK;')
        self.cur.execute('BEGIN;')

    def test_if_table_exists(
        self,
        table_name,
        schema_name,
        database_name=None,
    ):
        self.execute('USE DATABASE {0}'.format(
            database_name or self.database_name,
        ))
        return 0 < len(self.execute_and_return_result(
            """
            SELECT * FROM "{0}".information_schema.tables
            WHERE table_schema LIKE '{1}'
            AND table_name LIKE '{2}'
            """.format(
                database_name or self.database_name,
                schema_name,
                table_name,
            )
        ))



"""
    A utility to use SSH tunnel port forwarding including SSH agent forwarding.
    The motivation is that existing packages either do SSH tunnel forwarding
    well (`sshtnnel.SSHTunnelForwarder`) or they do SSH agent forwarding
    (`paramiko`), but no out of the box solution can provide both. EWAH
    uses this utility to implement SSH tunneling with and without SSH agent
    forwarding across all operators.

    Inspiration for the utility came from:
    - This stack overflow post: https://stackoverflow.com/questions/23666600/ssh-key-forwarding-using-python-paramiko
    - This implementation of SSH port forwarding: https://github.com/paramiko/paramiko/blob/master/demos/forward.py
    - This implementation of SSH agent forwarding: https://gist.github.com/toejough/436540622530c35404e6
"""

from airflow.hooks.base_hook import BaseHook

import getpass
import os
import socket
import select
import socketserver
import sys
import paramiko
import threading

from tempfile import NamedTemporaryFile

import sshtunnel

# The default timeout of 10s is too little if airflow runs many EL tasks in
# parallel -> if the SSH connection attempts are done in parallel, the timeout
# hits before the connection is established.
sshtunnel.TUNNEL_TIMEOUT = 30


def start_ssh_tunnel(ssh_conn_id, remote_conn_id, custom_timeout=None):
    """
    Start the SSH Tunnel with appropriate port forwarding.

    Returns both the SSHTunnelForwarder object for later closure and an
    adjusted connection object (host and port are adjusted in accordance with
    the open tunnel)
    """
    if custom_timeout:
        sshtunnel.TUNNEL_TIMEOUT = custom_timeout
    ssh_conn = BaseHook.get_connection(ssh_conn_id)
    remote_conn = BaseHook.get_connection(remote_conn_id)
    kwargs = {
        'ssh_address_or_host': (ssh_conn.host, ssh_conn.port or 22),
        'remote_bind_address': (remote_conn.host, remote_conn.port),
    }
    if ssh_conn.login:
        kwargs['ssh_username'] = ssh_conn.login or None
    if ssh_conn.password:
        kwargs['ssh_password'] = ssh_conn.password

    with NamedTemporaryFile() as keyfile:
        extra = ssh_conn.extra
        if extra:
            keyfile.write(extra.encode())
            keyfile.flush()
            kwargs['ssh_pkey'] = os.path.abspath(keyfile.name)
        ssh_tunnel_forwarder = sshtunnel.SSHTunnelForwarder(**kwargs)
        ssh_tunnel_forwarder.start()

    remote_conn.host = 'localhost'
    remote_conn.port = ssh_tunnel_forwarder.local_bind_port

    return ssh_tunnel_forwarder, remote_conn



from ewah.operators.base_operator import EWAHBaseOperator
from ewah.ewah_utils.airflow_utils import airflow_datetime_adjustments
from ewah.constants import EWAHConstants as EC

from airflow.hooks.base_hook import BaseHook

import gspread
import json
from oauth2client.service_account import ServiceAccountCredentials as SAC

class EWAHGSpreadOperator(EWAHBaseOperator):

    _IS_INCREMENTAL = False
    _IS_FULL_REFRESH = True

    _REQUIRES_COLUMNS_DEFINITION = True

    def _translate_alphanumeric_column(self, column_identifier):
        if type(column_identifier) == str:
            column_number = 0
            i = 0
            ident_dict = {}
            while column_identifier:
                letter = column_identifier[-1:].lower()
                if ord(letter) > ord('z') or ord(letter) < ord('a'):
                    raise Exception('Column letter {0} out of bounds!'.format(
                        letter,
                    ))
                column_identifier = column_identifier[:-1]
                ident_dict.update({i: ord(letter) + 1 - ord('a')})
                i += 1
            return sum([v * (26 ** k) for k, v in ident_dict.items()])
        else:
            return column_identifier

    def __init__(
        self,
        workbook_key, # can be seen in the URL of the workbook
        sheet_key, # name of the worksheet
        start_row=2, # in what row does the data begin?
        end_row=None, # optional: what is the last row? None gets all data
    *args, **kwargs):
        super().__init__(*args, **kwargs)

        credentials = self.source_conn.extra_dejson
        if not credentials.get('client_secrets'):
            raise Exception('Google Analytics Credentials misspecified!' \
                + ' Example of a correct specifidation: {0}'.format(
                    json.dumps({"client_secrets":{
                        "type": "service_account",
                        "project_id": "abc-123",
                        "private_key_id": "123456abcder",
                        "private_key": "-----BEGIN PRIVATE KEY-----\nxxx\n-----END PRIVATE KEY-----\n",
                        "client_email": "xyz@abc-123.iam.gserviceaccount.com",
                        "client_id": "123457",
                        "auth_uri": "https://accounts.google.com/o/oauth2/auth",
                        "token_uri": "https://oauth2.googleapis.com/token",
                        "auth_provider_x509_cert_url": "https://www.googleapis.com/oauth2/v1/certs",
                        "client_x509_cert_url": "https://www.googleapis.com/robot/v1/metadata/x509/xyz%40abc-123.iam.gserviceaccount.com"
                    }})
                ))

        column_match = {}
        for col_key, col_def in self.columns_definition.items():
            if (not col_def) or (not col_def.get(EC.QBC_FIELD_GSHEET_COLNO)):
                raise Exception((
                        'Column {0} is missing information regarding the ' \
                        + 'position of the column in the sheet.'
                    ).format(col_key)
                )
            column_match.update({
                self._translate_alphanumeric_column(
                    col_def[EC.QBC_FIELD_GSHEET_COLNO],
                ): col_key,
            })

        self.client_secrets = credentials['client_secrets']
        self.column_match = column_match
        self.workbook_key = workbook_key
        self.sheet_key = sheet_key
        self.start_row = start_row
        self.end_row = end_row

    def ewah_execute(self, context):
        client = gspread.authorize(
            SAC.from_json_keyfile_dict(
                self.client_secrets,
                ['https://spreadsheets.google.com/feeds'],
            ),
        )

        self.log.info('Retrieving data...')
        workbook = client.open_by_key(self.workbook_key)
        sheet = workbook.worksheet(self.sheet_key)
        raw_data = sheet.get_all_values()[self.start_row-1:self.end_row]

        # Load the data from the sheet into a format for upload into the DWH
        data = []
        for row in raw_data: # Iterate through each row
            data_dict = {} # New row, new dictionary of field:value
            row_is_null = True # Is entire row empty?
            for position, column in self.column_match.items():
                datapoint = row[position-1]
                data_dict.update({column:datapoint})
                if row_is_null:
                    if bool(datapoint):
                        if not (datapoint == '0'):
                            # Special case with zeroes in text format
                            row_is_null = False # Row is not empty!
            if not row_is_null: # Ignore empty rows
                data.append(data_dict)

        self.upload_data(data)



try: # load operators with dependencies installed, ignore others
    from ewah.operators.fx_operator import EWAHFXOperator
except ImportError:
    from ewah.operators.base_operator import EWAHEmptyOperator as EWAHFXOperator
try:
    from ewah.operators.google_analytics_operator import EWAHGAOperator
except ImportError:
    from ewah.operators.base_operator import EWAHEmptyOperator as EWAHGAOperator
try:
    from ewah.operators.mysql_operator import EWAHMySQLOperator
except ImportError:
    from ewah.operators.base_operator import EWAHEmptyOperator as EWAHMySQLOperator
try:
    from ewah.operators.oracle_operator import EWAHOracleSQLOperator
except ImportError:
    from ewah.operators.base_operator import EWAHEmptyOperator as EWAHOracleSQLOperator
try:
    from ewah.operators.postgres_operator import EWAHPostgresOperator
except ImportError:
    from ewah.operators.base_operator import EWAHEmptyOperator as EWAHPostgresOperator
try:
    from ewah.operators.s3_operator import EWAHS3Operator
except ImportError:
    from ewah.operators.base_operator import EWAHEmptyOperator as EWAHS3Operator
try:
    from ewah.operators.facebook_operator import EWAHFBOperator
except ImportError:
    from ewah.operators.base_operator import EWAHEmptyOperator as EWAHFBOperator
try:
    from ewah.operators.google_sheets_operator import EWAHGSpreadOperator
except ImportError:
    from ewah.operators.base_operator import EWAHEmptyOperator as EWAHGSpreadOperator
try:
    from ewah.operators.mongodb_operator import EWAHMongoDBOperator
except ImportError:
    from ewah.operators.base_operator import EWAHEmptyOperator as EWAHGSpreadOperator
try:
    from ewah.operators.shopify_operator import EWAHShopifyOperator
except ImportError:
    from ewah.operators.base_operator import EWAHEmptyOperator as EWAHShopifyOperator
try:
    from ewah.operators.zendesk_operator import EWAHZendeskOperator
except ImportError:
    from ewah.operators.base_operator import EWAHEmptyOperator as EWAHZendeskOperator
try:
    from ewah.operators.google_ads_operator import EWAHGoogleAdsOperator
except ImportError:
    from ewah.operators.base_operator import EWAHEmptyOperator as EWAHGoogleAdsOperator
try:
    from ewah.operators.google_maps_operator import EWAHGMapsOperator
except ImportError:
    from ewah.operators.base_operator import EWAHEmptyOperator as EWAHGMapsOperator
try:
    from ewah.operators.mailchimp_operator import EWAHMailchimpOperator
except ImportError:
    from ewah.operators.base_operator import EWAHEmptyOperator as EWAHMailchimpOperator
try:
    from ewah.operators.bigquery_operator import EWAHBigQueryOperator
except ImportError:
    from ewah.operators.base_operator import EWAHEmptyOperator as EWAHBigQueryOperator
try:
    from ewah.operators.hubspot_operator import EWAHHubspotOperator
except ImportError:
    from ewah.operators.base_operator import EWAHEmptyOperator as EWAHHubspotOperator
try:
    from ewah.operators.stripe_operator import EWAHStripeOperator
except ImportError:
    from ewah.operators.base_operator import EWAHEmptyOperator as EWAHStripeOperator



from ewah.operators.base_operator import EWAHBaseOperator
from ewah.constants import EWAHConstants as EC

from airflow.hooks.base_hook import BaseHook

import requests
import json
import urllib
import time


class EWAHHubspotOperator(EWAHBaseOperator):

    _IS_INCREMENTAL = False
    _IS_FULL_REFRESH = True

    _REQUIRES_COLUMNS_DEFINITION = False

    _BASE_URL = 'https://api.hubapi.com/crm/v3/objects/{0}'
    _PROPERTIES_URL = 'https://api.hubapi.com/crm/v3/properties/{0}'

    _OBJECTS = [
        'companies',
        'contacts',
        'deals',
        'feedback_submissions',
        'line_items',
        'products',
        'tickets',
        'quotes',
    ]

    def __init__(self, *args, object=None, properties=None, **kwargs):
        if object is None:
            object = kwargs.get('target_table_name')
        assert object in self._OBJECTS, 'Object {0} is invalid!'.format(object)
        kwargs['primary_key_column_name'] = 'id'
        super().__init__(*args, **kwargs)
        self.object = object
        self.properties = properties

    def ewah_execute(self, context):
        params = {'hapikey': self.source_conn.password, 'limit': 100}
        headers = {'accept': 'application/json'}
        url = self._BASE_URL.format(self.object)
        url_properties = self._PROPERTIES_URL.format(self.object)

        # set properties list
        properties = self.properties
        if properties is None:
            # get properties via API call if none are given
            request = requests.get(
                url=url_properties,
                params=params,
                headers=headers,
            )
            assert request.status_code == 200, request.text
            result = json.loads(request.text)['results']
            properties = [property['name'] for property in result]
        params.update({'properties': properties})

        # get and upload data
        keepgoing = True
        counter = 0
        while keepgoing:
            time.sleep(0.2)
            counter += 1
            _msg = 'Loading data - request number {0}...'.format(str(counter))
            self.log.info(_msg)
            request = requests.get(url=url, params=params, headers=headers)
            if request.status_code == 414:
                _msg = 'ERROR! Too many properties '
                _msg += '- please use a smaller, custom set of properties!'
                _msg += '\n\navailable properties:\n\n\t'
                _msg += '\n\t'.join(properties)
                self.log.info(_msg)
            assert request.status_code == 200, request.text
            response = json.loads(request.text)
            # keep going as long as a link is shipped in the response
            keepgoing = response.get('paging', {}).get('next', {}).get('after')
            params.update({'after': keepgoing})
            self.upload_data(response['results'])



from ewah.operators.base_operator import EWAHBaseOperator
from ewah.ewah_utils.airflow_utils import airflow_datetime_adjustments
from ewah.ewah_utils.python_utils import is_iterable_not_string
from ewah.constants import EWAHConstants as EC

from google.ads.google_ads.client import GoogleAdsClient
from airflow.hooks.base_hook import BaseHook

from datetime import datetime, timedelta
from copy import deepcopy

class EWAHGoogleAdsOperator(EWAHBaseOperator):

    template_fields = ('data_from', 'data_until')

    _IS_INCREMENTAL = True
    _IS_FULL_REFRESH = False

    _REQUIRED_KEYS = (
        'developer_token',
        'client_id',
        'client_secret',
        'refresh_token',
    )

    def __init__(self,
        fields, # dict of fields, e.g. {'campaign': ['id', 'name']}
        metrics, # list of metrics to retrieve -
        resource, # name of the resource, e.g. ad_group
        client_id,
        conditions=None, # list of conditions, e.g. ["ad_group.status = 'ENABLED'"]
        data_from=None, # can be datetime or timedelta relative to data_until
        data_until=None,
    *args, **kwargs):
        """
            recommendation: leave data_until=None and use a timedelta for
            data_from.
        """

        if kwargs.get('columns_definition'):
            raise Exception('columns_definition is not accepted for this ' \
                + 'operator!')

        kwargs['update_on_columns'] = [
            col.replace('.', '_') for col in self.get_select_statement(fields)
            if col[:7] == 'segment'
        ] + ['{0}_resource_name'.format(resource)]

        if conditions and not is_iterable_not_string(conditions):
            raise Exception('Argument "conditions" must be a list!')

        if not (fields.get('segments') and ('date' in fields['segments'])):
            raise Exception('fields list MUST contain segments.date!')

        # must use deepcopy due to airflow calling field by reference
        # and the update in the following row adding the metrics to the field,
        # which in turn screws with the update_on_columns definition above
        self.fields_dict = deepcopy(fields)
        self.fields_dict.update({'metrics': deepcopy(metrics)})
        self.fields_list = self.get_select_statement(self.fields_dict)
        self.resource = resource
        self.client_id = str(client_id)
        self.conditions = conditions
        self.data_from = data_from
        self.data_until = data_until

        super().__init__(*args, **kwargs)

    def get_select_statement(self, dict_format, prefix=None):
        # create the list of fields for the SELECT statement
        if prefix is None:
            prefix = ''
        elif not prefix[-1] == '.':
            prefix += '.'
        fields = []
        for key, value in dict_format.items():
            for item in value:
                if type(item) == dict:
                    fields += self.get_select_statement(item, prefix + key)
                else:
                    fields += [prefix + key + '.' + item]
        return fields

    def ewah_execute(self, context):
        # Task execution happens here
        def get_data_from_ads_output(fields_dict, values, prefix=None):
            if prefix is None:
                prefix = ''
            elif not prefix[-1] == '_':
                prefix += '_'
                # e.g. 2b prefix = 'ad_group_criterion_'
            data = {}
            for key, value in fields_dict.items():
                # e.g. 1 key = 'metrics', value = ['impressions', 'clicks']
                # e.g. 2a key = 'ad_group_criterion', value = [{'keyword': ['text', 'match_type']}]
                # e.g. 2b key = 'keyword', value = ['text', 'match_type']
                node = getattr(values, key)
                # e.g. 1 node = row.metrics
                # e.g. 2a node = row.ad_group_criterion
                # e.g. 2b node = row.ad_group_criterion.keyword
                for item in value:
                    # e.g. 1 item = 'clicks'
                    # e.g. 2a item = {'keyword': ['text', 'match_type']}
                    # e.g. 2b item = 'text'
                    if type(item) == dict:
                        data.update(get_data_from_ads_output(
                            fields_dict=item,
                            values=node,
                            prefix=prefix + key, # e.g. 2a '' + 'ad_group_criterion'
                        ))
                    else:
                        # e.g. 1: {'' + 'metrics' + '_' + 'clicks': row.metrics.clicks.value}
                        # e.g. 2b: {'ad_group_criterion_' + 'keyeword' + '_' + 'text': row.ad_group_criterion.keyword.text.value}
                        if hasattr(getattr(node, item), 'value'):
                            data.update({
                                prefix + key + '_' + item: \
                                    getattr(node, item).value
                            })
                        else:
                            # some node ends don't respond to .value but are
                            #   already the value
                            data.update({
                                prefix + key + '_' + item: getattr(node, item)
                            })
            return data


        self.data_until = airflow_datetime_adjustments(self.data_until)
        self.data_until = self.data_until or context['next_execution_date']
        if isinstance(self.data_from, timedelta):
            self.data_from = self.data_until - self.data_from
        else:
            self.data_from = airflow_datetime_adjustments(self.data_from)
            self.data_from = self.data_from or context['execution_date']

        conn = self.source_conn.extra_dejson
        credentials = {}
        for key in self._REQUIRED_KEYS:
            if not key in conn.keys():
                raise Exception('{0} must be in connection extra json!'.format(
                    key
                ))
            credentials[key] = conn[key]

        # build the query
        query = 'SELECT {0} FROM {1} WHERE segments.date {2} {3}'.format(
            ', '.join(self.fields_list),
            self.resource,
            "BETWEEN '{0}' AND '{1}'".format(
                self.data_from.strftime('%Y-%m-%d'),
                self.data_until.strftime('%Y-%m-%d'),
            ),
            ('AND' + ' AND '.join(self.conditions)) if self.conditions else '',
        )

        self.log.info('executing this google ads query:\n{0}'.format(query))
        cli = GoogleAdsClient.load_from_dict(credentials)
        service = cli.get_service("GoogleAdsService", version="v3")
        search = service.search(
            self.client_id.replace('-', ''),
            query=query,
        )
        data = [row for row in search]

        # get into uploadable format
        upload_data = []
        fields_dict = deepcopy(self.fields_dict)
        fields_dict.update({self.resource: ['resource_name']})
        while data:
            datum = data.pop(0)
            upload_data += [get_data_from_ads_output(
                fields_dict,
                datum,
            )]

        self.upload_data(upload_data)



from airflow.hooks.base_hook import BaseHook

from ewah.constants import EWAHConstants as EC

import json
import hashlib
from copy import deepcopy
from collections import OrderedDict
from psycopg2.extras import RealDictCursor

class EWAHBaseDWHook(BaseHook):
    """Extension of airflow's native Base Hook.

    Contains a few functions used by the EWAH Hooks for data cleaning prior
    to uploading. Each implemented DWH Engine has its own Ewah hook, which is
    always a child of this class and contains logic that is DWH specific.
    """

    def __init__(self, dwh_engine, dwh_conn, *args, logging_func=None, **kwarg):
        args = [dwh_conn.conn_id] + (args or [])
        super().__init__(*args, **kwarg)
        self.dwh_engine = dwh_engine
        self.credentials = dwh_conn
        self.logging_func = logging_func or print
        self._init_conn(first_call=True)

    def __del__(self, *args, **kwargs):
        self.close()
        if hasattr(super(), '__del__'):
            super().__del__(*args, **kwargs)

    def _init_conn(self, first_call=False, commit=False):
        if not first_call:
            if commit:
                self.commit()
            self.close()
        self.conn = self._create_conn()
        self.cur = self.conn.cursor()

        if self.dwh_engine == EC.DWH_ENGINE_POSTGRES:
            self.dict_cur = self.conn.cursor(
                cursor_factory=RealDictCursor
            )

        if self.dwh_engine == EC.DWH_ENGINE_SNOWFLAKE:
            self.cur.execute("BEGIN;")

    def _get_column_type(self, columns_definition):
        return columns_definition.get(
            EC.QBC_FIELD_TYPE,
            EC.QBC_TYPE_MAPPING[self.dwh_engine].get(
                EC.QBC_TYPE_MAPPING_DEFAULT,
                EC.QBC_TYPE_MAPPING_INCONSISTENT,
            ),
        )

    def close(self):
        if hasattr(self, 'cur') and self.cur:
            self.cur.close()
        if hasattr(self, 'dict_cur') and self.dict_cur:
            self.dict_cur.close()
        if self.conn:
            self.conn.close()

    def execute(self, sql, params=None, commit=False, cursor=None):
        self.logging_func('\nExecuting SQL:\n\n{0}'.format(sql))
        kwargs = {}
        args = []
        if params:
            if self.dwh_engine == EC.DWH_ENGINE_POSTGRES:
                kwargs.update({'vars': params})
            elif self.dwh_engine == EC.DWH_ENGINE_SNOWFLAKE:
                args = [params]
            else:
                raise Exception('Feature not implemented!')

        cur = cursor or self.cur
        if self.dwh_engine == EC.DWH_ENGINE_SNOWFLAKE:
            # Snowflake does not allow multiple statements in one call
            # refactor this! breaks queries with strings including semicolon!
            for statement in sql.strip().split(';'):
                if statement.strip():
                    cur.execute(statement.strip(), *args, **kwargs)
        else:
            cur.execute(sql.strip(), *args, **kwargs)

        if commit:
            self.commit()

    def execute_and_return_result(self, sql, params=None, return_dict=False):
        if ';' in sql.strip()[:-1]:
            raise Exception(
                'This function only executes a single statement! This query'
                + ' consists of more than one statement:\n' + sql
            )

        self.execute(
            sql=sql,
            params=params,
            commit=False,
            cursor=self.dict_cur if return_dict else None,
        )

        if self.dwh_engine == EC.DWH_ENGINE_POSTGRES:
            if return_dict:
                return self.dict_cur.fetchall()
            return self.cur.fetchall()

        if self.dwh_engine == EC.DWH_ENGINE_SNOWFLAKE:
            return [row for row in self.cur]

        raise Exception(
            'Function not implemented for this DWH Engine: {0}'.format(
                self.dwh_engine,
            )
        )

    def get_cursor(self, dict_cursor=False):
        if dict_cursor:
            if not hasattr(self, 'dict_cur'):
                raise Exception(
                    'Dictionary cursor is not yet implemented for {0}!'
                    .format(self.dwh_engine)
                )
            return self.dict_cur
        return self.cur

    def detect_and_apply_schema_changes(
        self,
        new_schema_name,
        new_table_name,
        new_columns_dictionary,
        drop_missing_columns=False,
        database=None,
        commit=False,
    ):
        params = {
            'schema_name': new_schema_name,
            'table_name': new_table_name,
        }
        if self.dwh_engine == EC.DWH_ENGINE_SNOWFLAKE:
            database = database or self.database
            params.update({'database_name': database})
        if not self.test_if_table_exists(**params):
            return ([], []) # Table did not previously exist, so there is nothing to do

        list_of_old_columns = [
            col[0].strip() for col in self.execute_and_return_result(
                sql=self._QUERY_SCHEMA_CHANGES_COLUMNS.format(**params),
                params=params,
                return_dict=False,
            )]
        list_of_columns = [
            col_name.strip() for col_name in list(new_columns_dictionary.keys())
        ]

        dropped_columns = []
        if drop_missing_columns:
            for old_column in list_of_old_columns:
                if not (old_column in list_of_columns):
                    drop_params = deepcopy(params)
                    drop_params.update({'column_name': old_column})
                    dropped_columns += [old_column]
                    self.execute(
                        sql=self._QUERY_SCHEMA_CHANGES_DROP_COLUMN.format(
                            **drop_params,
                        ),
                        commit=False,
                    )

        new_columns = []
        for column in list_of_columns:
            if not (column in list_of_old_columns):
                new_columns += [column]
                add_params = deepcopy(params)
                add_params.update({
                    'column_name': column,
                    'column_type': self._get_column_type(
                        new_columns_dictionary[column],
                    )
                })
                self.execute(
                    sql=self._QUERY_SCHEMA_CHANGES_ADD_COLUMN.format(
                        **add_params,
                    ),
                    commit=False,
                )

        if commit:
            self.commit()

        return (new_columns, dropped_columns)

    def create_or_update_table(
        self,
        data,
        columns_definition,
        table_name,
        schema_name,
        schema_suffix,
        database_name=None,
        drop_and_replace=True,
        update_on_columns=None,
        commit=False,
        logging_function=None,
        clean_data_before_upload=True,
        hash_columns=None,
        hashlib_func_name=None,
    ):
        logging_function = logging_function or (lambda *args: None)
        database_name = database_name or \
            (getattr(self, 'database') if hasattr(self, 'database') else None)

        raw_row = {} # Used as template for params at execution
        sql_part_columns = [] # Used for CREATE and INSERT / UPDATE query
        jsonb_columns = [] # JSON columns require special treatment
        create_update_on_columns = (not (drop_and_replace or update_on_columns))
        update_on_columns = update_on_columns or []
        if not (type([]) == type (update_on_columns)):
            raise Exception('"update_on_columns" must be a list!' + \
                'Is currently: type {0}'.format(str(type(update_on_columns))))
        field_constraints_mapping = EC.QBC_FIELD_CONSTRAINTS_MAPPING.get(
            self.dwh_engine,
        )

        # don't add primary key to create table definition
        # instead make alter table call later for the case of composite PKs
        field_constraints_mapping.pop(EC.QBC_FIELD_PK, None)
        pk_columns = []

        hash_columns = hash_columns or []
        for column_name in columns_definition.keys():
            raw_row.update({column_name: None})
            definition = columns_definition[column_name]
            if not type({}) == type(definition):
                raise Exception('Column {0} is not properly defined!'.format(
                    column_name,
                ))
            sql_part_columns += ['"{0}"\t{1}\t{2}'.format(
                column_name, # Field name
                self._get_column_type(definition), # Type: text, int etc.
                ' '.join([ # Get all additional field properties (unique etc.)
                    field_constraints_mapping[addon] if definition.get(addon) else ''
                    for addon in list(field_constraints_mapping.keys())
                ])
            )]
            if self._get_column_type(definition) == 'jsonb':
                jsonb_columns += [column_name]
            if definition.get(EC.QBC_FIELD_PK):
                pk_columns += [column_name]
                if create_update_on_columns:
                    update_on_columns += [column_name]
            if definition.get(EC.QBC_FIELD_HASH):
                hash_columns += [column_name]
        sql_part_columns = ',\n\t'.join(sql_part_columns)

        if clean_data_before_upload:
            logging_function('Cleaning data for upload...')
            upload_data = []
            cols_list = list(raw_row.keys())
            if hash_columns:
                hash_func = getattr(hashlib, hashlib_func_name)
            while data:
                datum = data.pop(0)
                # Make sure that each dict in upload_data has all keys
                row = deepcopy(raw_row)
                for column_name, value in datum.items():
                    if column_name in cols_list:
                        if column_name in hash_columns:
                            # hash column!
                            pre_digest = hash_func(str(value).encode())
                            row[column_name] = pre_digest.hexdigest()
                        else:
                            # avoid edge case of data where all instances of a field
                            #   are None, thus having data for a field missing
                            #   in the columns_definition!
                            value_type = type(value)
                            if column_name in jsonb_columns and value:
                                row[column_name] = json.dumps(value)
                            elif value_type in [dict, OrderedDict, list]:
                                row[column_name] = json.dumps(value)
                            elif not (value == '\0'):
                                if value_type == str:
                                    row[column_name] = value.replace('\x00', '')
                                else:
                                    row[column_name] = value
                upload_data += [row]
            data_len = len(upload_data)
        else:
            data_len = len(data)
            upload_data = None

        logging_function('Uploading {0} rows of data...'.format(
            str(data_len),
        ))
        if database_name:
            kwargs = {'database_name': database_name}
        else:
            kwargs = {}
        kwargs.update({
            'data': upload_data or data,
            'table_name': table_name,
            'schema_name': schema_name,
            'schema_suffix': schema_suffix,
            'columns_definition': columns_definition,
            'columns_partial_query': sql_part_columns,
            'update_on_columns': update_on_columns,
            'drop_and_replace': drop_and_replace,
            'logging_function': logging_function,
            'pk_columns': pk_columns,
        })
        self._create_or_update_table(**kwargs)

        if commit:
            self.commit()



from airflow.models import BaseOperator
from airflow.hooks.base_hook import BaseHook
from airflow.utils.file import TemporaryDirectory

from ewah.constants import EWAHConstants as EC
from ewah.ewah_utils.yml_loader import Loader, Dumper
from ewah.ewah_utils.ssh_tunnel import start_ssh_tunnel

import re
import os
import venv
import yaml
import subprocess
from tempfile import NamedTemporaryFile

class EWAHdbtOperator(BaseOperator):
    """
        Runs dbt tasks, given a git repository and a target data warehouse.
    """

    # allowed commands (sans flags)
    _commands = [
        'run',
        'test',
        'snapshot',
        'seed',
        'docs generate',
        'docs serve',
    ]
    # note: dbt deps is always executed first!

    def __init__(self,
        repo_type,
        dwh_engine,
        dwh_conn_id,
        git_conn_id=None,
        dbt_commands=['run'], # string or list of strings - dbt commands
        dbt_version='0.18.1',
        git_link=None, # SSH link to a remote git repository, if using git repo
        git_branch=None, # optional: branch to check out
        subfolder=None, # optional: supply if dbt project is in a subfolder
        threads=4, # see https://docs.getdbt.com/dbt-cli/configure-your-profile/#understanding-threads
        schema_name='analytics', # see https://docs.getdbt.com/dbt-cli/configure-your-profile/#understanding-target-schemas
        database_name=None, # Snowflake only - name of the database
        keepalives_idle=0, # see https://docs.getdbt.com/reference/warehouse-profiles/postgres-profile/
        ssh_tunnel_id=None,
    *args, **kwargs):

        assert repo_type in ('git')
        assert dbt_commands
        assert dbt_version
        assert threads
        assert schema_name

        assert dwh_engine in (EC.DWH_ENGINE_POSTGRES, EC.DWH_ENGINE_SNOWFLAKE)

        # only Snowflake is allowed to have the database_name kwarg
        assert (dwh_engine == EC.DWH_ENGINE_SNOWFLAKE) or (not database_name)


        if isinstance(dbt_commands, str):
            dbt_commands = [dbt_commands]

        # check if all commands are allowed -- note that flags are allowed!
        if min([
                max([1 if cmd.startswith(_c) else 0 for _c in self._commands])
                for cmd in dbt_commands
            ]) == 0:
            raise Exception('dbt commands must be one of:\n\n\t{0}{1}'.format(
                '\n\t'.join(self._commands),
                '\n\ncurrently: {0}'.format(dbt_commands),
            ))

        if repo_type == 'git':
            assert git_link
            assert re.search('@(.*):', git_link) # only SSH links currently work

        super().__init__(*args, **kwargs)

        self.repo_type = repo_type
        self.git_conn_id = git_conn_id
        self.dwh_engine = dwh_engine
        self.dwh_conn_id = dwh_conn_id
        self.dbt_commands = dbt_commands
        self.dbt_version = dbt_version
        self.git_link = git_link
        self.git_branch = git_branch
        self.subfolder = subfolder
        self.threads = threads
        self.schema_name = schema_name
        self.keepalives_idle = keepalives_idle
        self.ssh_tunnel_id = ssh_tunnel_id
        self.database_name = database_name

    def execute(self, context):
        def run_cmd(cmd, env):
            # run a bash command (cmd) with the environment variables env
            self.log.info('running:\n\n\t{0}\n\n'.format(cmd))
            with subprocess.Popen(
                cmd,
                executable='/bin/bash',
                shell=True, # required when parsing string as command
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                env=env,
            ) as p:
                # write output to log without delay
                for stdout_line in p.stdout:
                    self.log.info(stdout_line.decode().strip())
            assert p.returncode == 0 # fail task if process failed

        # env to be used in processes later
        env = os.environ.copy()

        # create a new temp folder, all action happens in here
        with TemporaryDirectory(prefix='__ewah_dbt_operator_') as tmp_dir:
            # clone repo into temp directory
            repo_dir = tmp_dir + os.path.sep + 'repo'
            if self.repo_type == 'git':
                git_repo_link = ''
                if self.git_branch:
                    git_repo_link = '-b {0} '.format(self.git_branch)
                git_repo_link += self.git_link
                with NamedTemporaryFile(dir=tmp_dir) as f:
                    # temporarily have a file with the SSH key when cloning
                    cmd = 'eval `ssh-agent`'
                    if self.git_conn_id:
                        git_conn = BaseHook.get_connection(self.git_conn_id)
                        ssh_key_text = git_conn.extra
                        file_name = os.path.abspath(f.name)
                        f.write(ssh_key_text.encode())
                        f.seek(0)
                        # if there is a password set in the git conn, it's
                        # the password for the SSH key! supply it!
                        # use SSH_ASKPASS env var & a temp file to do so
                        ssh_pw = ''
                        if git_conn.password:
                            f_pw_path = tmp_dir + os.path.sep + 'ssh_key_pw'
                            with open(f_pw_path, 'w+') as f_pw:
                                # create file in temp folder -> ssh-add cannot
                                # open a NamedTemporaryFile in context manager
                                f_pw.write('#!/bin/bash\necho "{0}"'
                                        .format(git_conn.password)
                                )
                                f_pw.seek(0)
                            ssh_pw = 'DISPLAY=":0.0" SSH_ASKPASS="{0}" '
                            ssh_pw = ssh_pw.format(f_pw_path)
                            cmd += ' && chmod 777 {0}'.format(f_pw_path)
                        cmd += ' && {1}ssh-add {0}'.format(file_name, ssh_pw)
                    else:
                        cmd += 'echo "cloning without credentials!"'
                    ssh_domain = re.search('@(.*):', self.git_link).group(1)
                    cmd += ' && mkdir -p $HOME/.ssh'
                    cmd += (' && ssh-keyscan -H {0} >> $HOME/.ssh/known_hosts'
                    .format(ssh_domain))
                    cmd += ' && git clone {0} {1}'.format(
                        git_repo_link,
                        repo_dir,
                    )
                    run_cmd(cmd, env) # cloning repo
            else:
                raise Exception('Not Implemented!')

            # create a virual environment in temp folder
            venv_folder = tmp_dir + os.path.sep + 'venv'
            self.log.info('creating a new virtual environment in {0}...'.format(
                venv_folder,
            ))
            venv.create(venv_folder, with_pip=True)

            # install dbt into created venv
            self.log.info('installing dbt=={0}'.format(self.dbt_version))
            cmd = 'source {0}/bin/activate'.format(venv_folder)
            cmd += (' && pip install --quiet --upgrade dbt=={0}'
                    .format(self.dbt_version))
            cmd += ' && dbt --version'
            cmd += ' && deactivate'
            run_cmd(cmd, env)

            dbt_dir = repo_dir
            if self.subfolder:
                if not self.subfolder[:1] == os.path.sep:
                    self.subfolder = os.path.sep + self.subfolder
                dbt_dir += self.subfolder

            # if applicable: open SSH tunnel
            if self.ssh_tunnel_id:
                self.ssh_tunnel_forwarder, dwh_conn = start_ssh_tunnel(
                    ssh_conn_id=self.ssh_tunnel_id,
                    remote_conn_id=self.dwh_conn_id,
                )
            else:
                dwh_conn = BaseHook.get_connection(self.dwh_conn_id)


            # read profile name & create temporary profiles.yml
            project_yml_file = dbt_dir
            if not project_yml_file[-1:] == os.path.sep:
                project_yml_file += os.path.sep
            project_yml_file += 'dbt_project.yml'
            project_yml = yaml.load(open(project_yml_file, 'r'), Loader=Loader)
            profile_name = project_yml['profile']
            self.log.info('Creating temp profile "{0}"'.format(profile_name))
            profiles_yml = {
                'config': {
                    'send_anonymous_usage_stats': False,
                    'use_colors': False, # colors won't be useful in logs
                },
            }
            if self.dwh_engine == EC.DWH_ENGINE_POSTGRES:
                profiles_yml[profile_name] = {
                    'target': 'prod', # same as the output defined below
                    'outputs': {
                        'prod': { # for postgres
                            'type': 'postgres',
                            'host': dwh_conn.host,
                            'port': dwh_conn.port or '5432',
                            'user': dwh_conn.login,
                            'pass': dwh_conn.password,
                            'dbname': dwh_conn.schema,
                            'schema': self.schema_name,
                            'threads': self.threads,
                            'keepalives_idle': self.keepalives_idle,
                        },
                    },
                }
            elif self.dwh_engine == EC.DWH_ENGINE_SNOWFLAKE:
                extra = dwh_conn.extra_dejson
                _db = extra.get('database', self.database_name)
                _db = _db or dwh_conn.schema
                profiles_yml[profile_name] = {
                    'target': 'prod', # same as the output defined below
                    'outputs': {
                        'prod': { # for postgres
                            'type': 'snowflake',
                            'account': extra.get('account', dwh_conn.host),
                            'user': extra.get('user', dwh_conn.login),
                            'password': extra.get('password', dwh_conn.password),
                            'role': extra.get('role'),
                            'database': _db,
                            'warehouse': extra.get('warehouse'),
                            'schema': self.schema_name,
                            'threads': self.threads,
                            'keepalives_idle': self.keepalives_idle,
                        },
                    },
                }
            else:
                raise Exception('DWH Engine not implemented!')

            # run commands with correct profile in the venv in the temp folder
            profiles_yml_name = tmp_dir + os.path.sep + 'profiles.yml'
            env['DBT_PROFILES_DIR'] = os.path.abspath(tmp_dir)
            with open(profiles_yml_name, 'w') as profiles_file:
                # write profile into profiles.yml file
                yaml.dump(profiles_yml, profiles_file, default_flow_style=False)

                # run dbt commands
                self.log.info('Now running commands dbt!')
                cmd = 'cd {0}'.format(dbt_dir)
                cmd += ' && source {0}/bin/activate'.format(venv_folder)
                cmd += ' && dbt '.join(['', 'deps'] + self.dbt_commands)
                cmd += ' && deactivate'
                run_cmd(cmd, env)

            # if applicable: close SSH tunnel
            if hasattr(self, 'ssh_tunnel_forwarder'):
                self.log.info('Stopping!')
                self.ssh_tunnel_forwarder.stop()
                del self.ssh_tunnel_forwarder



from ewah.operators.base_operator import EWAHBaseOperator
from ewah.ewah_utils.airflow_utils import airflow_datetime_adjustments
from ewah.ewah_utils.python_utils import is_iterable_not_string
from ewah.constants import EWAHConstants as EC

from airflow.hooks.base_hook import BaseHook

from datetime import datetime, timedelta
from pytz import timezone

from requests.auth import HTTPBasicAuth
import requests
import json
import time
import copy

class EWAHShopifyOperator(EWAHBaseOperator):

    # template_fields = ()

    _IS_INCREMENTAL = True
    _IS_FULL_REFRESH = False

    _acceptable_api_versions = [
        '2020-07',
    ]
    _current_api_version = '2020-07'
    _base_url = 'https://{shop}.myshopify.com/admin/api/{version}/{object}.json'

    # Set all accepted objects with their potential filter fields with defaults
    #   Note: always expected fields updated_at_max, updated_at_min, limit, id
    #   the key _is_drop_and_replace indicates objects that are only able
    #   to load with a full refresh
    _default_timestamp_fields = ('updated_at_min', 'updated_at_max')
    _default_datetime_format = '%Y-%m-%d %H:%M:%S%z'
    _accepted_objects = {
        'balance_transactions': {
            '_is_drop_and_replace': True,
            '_object_url': 'shopify_payments/balance/transactions',
            '_name_in_request_data': 'transactions',
            'since_id': None,
            'last_id': None,
            'test': None,
            'payout_id': None,
            'payout_status': None,
        },
        'customers': {
            'ids': None,
            'since_id': None,
        },
        'disputes': {
            '_is_drop_and_replace': True,
            '_object_url': 'shopify_payments/disputes',
        },
        'events': {
            '_timestamp_fields': ('created_at_min', 'created_at_max'),
            'since_id': None,
            'filter': None,
            'verb': None,
        },
        'orders': {
            'ids': None,
            'since_id': None,
            'status': 'any',
            'financial_status': None,
            'fulfillment_status': None,
            'fields': None,
        },
        'payouts': {
            '_timestamp_fields': ('date_min', 'date_max'),
            '_datetime_format': '%Y-%m-%d',
            '_object_url': 'shopify_payments/payouts',
            'since_id': None,
            'last_id': None,
            'status': None,
        },
        'products': {
            'ids': None,
            'since_id': None,
            'title': None,
            'vendor': None,
            'handle': None,
            'product_type': None,
            'collection_id': None,
            'published_status': None,
        },
        'tender_transactions': {
            '_timestamp_fields': ('processed_at_min', 'processed_at_max'),
            'since_id': None,
        },
    }

    def __init__(self,
        shop_id,
        shopify_object,
        auth_type,
        filter_fields={},
        api_version=None,
        get_transactions_with_orders=False,
        get_events_with_orders=False,
        get_inventory_data_with_product_variants=False,
        page_limit=250, # API Call pagination limit
    *args, **kwargs):

        if is_iterable_not_string(shop_id):
            raise Exception('Multiple shops in one DAG is deprecated!')

        if get_transactions_with_orders and not shopify_object == 'orders':
            raise Exception('transactions can only be pulled for orders!')

        if get_events_with_orders and not shopify_object == 'orders':
            raise Exception('events can only be pulled for orders!')

        if get_inventory_data_with_product_variants \
            and not shopify_object == 'products':
            raise Exception('inventory data may only be pulled with products!')

        if not shopify_object in self._accepted_objects.keys():
            raise Exception('{0} is not in the list of accepted objects!' + \
                ' accepted objects: {1}'.format(
                    shopify_object,
                    ', '.join(self._accepted_objects.keys())
                ))

        if self._accepted_objects[shopify_object].get('_is_drop_and_replace'):
            kwargs['drop_and_replace'] = True

        if not auth_type in ['access_token', 'basic_auth']:
            raise Exception('auth_type must be access_token or basic_auth!')

        if not type(filter_fields) == dict:
            raise Exception('filter_fields must be a dictionary!')
        else:
            for key, value in filter_fields.items():
                if not key in self._accepted_objects['shopify_object'].keys():
                    raise Exception('invalid key {0} in filter fields!'.format(
                        key,
                    ))

        api_version = api_version or self._current_api_version
        if not api_version in self._acceptable_api_versions:
            raise Exception(
                '{0} is not a valid api version! valid versions: {1}'.format(
                    api_version,
                    ', '.join(self._acceptable_api_versions),
                )
            )

        if not type(page_limit) == int or page_limit > 250 or page_limit < 1:
            raise Exception( \
                'Page limit must be a positive integer not exceeding 250!')

        kwargs['primary_key_column_name'] = \
            kwargs.get('primary_key_column_name', 'id')

        # source conn id is not required on operator call level! avoid error
        kwargs['source_conn_id'] = kwargs.get('source_conn_id', '__none__')

        super().__init__(*args, **kwargs)

        self.shop_id = shop_id
        self.shopify_object = shopify_object
        self.auth_type = auth_type
        self.filter_fields = filter_fields
        self.api_version = api_version
        self.page_limit = page_limit
        self.get_transactions_with_orders = get_transactions_with_orders
        self.get_events_with_orders = get_events_with_orders
        self.get_inventory_data_with_product_variants = \
            get_inventory_data_with_product_variants

    def ewah_execute(self, context):
        # can supply a list of shops - need to run for all shops individually!
        def datetime_to_string(dt, format):
            # check if tz aware; set to utc if so
            if dt.tzinfo:
                dt = dt.astimezone(timezone('UTC'))
            else:
                dt = dt.replace(tzinfo=timezone('UTC'))
            # check if format_string contains timezone
            if '%z' in format:
                # add colon!
                dt_string = dt.strftime(format)
                return dt_string[:-2] + ':' + dt_string[-2:]
            else:
                return dt.strftime(format)

        object_metadata = self._accepted_objects[self.shopify_object]
        self.object_metadata = object_metadata
        params = {
            key: val
            for key, val in object_metadata.items()
            if not val is None and not key[:1] == '_'
        }
        params.update(self.filter_fields)
        params.update({'limit': self.page_limit})
        if not self.drop_and_replace:
            timestamp_fields = object_metadata.get(
                '_timestamp_fields',
                self._default_timestamp_fields,
            )
            timestamp_format_string = object_metadata.get(
                '_datetime_format',
                self._default_datetime_format,
            )
            params.update({
                # Pendulum by coincidence converts to the correct string format
                timestamp_fields[1]: datetime_to_string(
                        context['next_execution_date'],
                        timestamp_format_string,
                    ),
            })
            if self.test_if_target_table_exists():
                params.update({
                    timestamp_fields[0]: datetime_to_string(
                        context['execution_date'],
                        timestamp_format_string,
                    ),
                })

        source_conn_id = self.source_conn.conn_id
        auth_type = self.auth_type
        if is_iterable_not_string(self.shop_id):
            # multiple shops to iterate - loop through!
            # deprecated feature - don't use!
            raise Exception('Multiple Shops in one DAG is deactivated!')
            self.log.info('iterating through multiple shops!')
            for shop_id in self.shop_id:
                # metadata: shop id
                self.log.info('getting data for: {0}'.format(shop_id))
                if hasattr(self.shop_id, 'get') \
                    and hasattr(self.shop_id[shop_id], 'get'):
                    # dict, not list! check for conn details!
                    shop_dict = self.shop_id[shop_id]
                    sci = shop_dict.get('source_conn_id', source_conn_id)
                    at = shop_dict.get('auth_type', auth_type)
                else:
                    sci = source_conn_id
                    at = auth_type
                self._metadata.update({'shop_id': shop_id})
                self.execute_for_shop(context, shop_id, params, sci, at)
        else:
            self._metadata.update({'shop_id': self.shop_id})
            sci = self.source_conn.conn_id
            at = self.auth_type
            self.execute_for_shop(context, self.shop_id, params, sci, at)

    def execute_for_shop(self,
        context,
        shop_id,
        params,
        source_conn_id,
        auth_type,
    ):
        # Get data from shopify via REST API
        def add_get_transactions(data, shop, version, req_kwargs):
            # workaround to add transactions to orders
            self.log.info('Requesting transactions of orders...')
            base_url = "https://{shop}.myshopify.com/admin/api/{version}/orders/{id}/transactions.json"
            base_url = base_url.format(**{
                'shop': shop,
                'version': version,
                'id': '{id}',
            })

            for datum in data:
                id = datum['id']
                # self.log.info('getting transactions for order {0}'.format(id))
                time.sleep(1) # avoid hitting api call requested per second limit
                url = base_url.format(id=id)
                req = requests.get(url, **req_kwargs)
                if not req.status_code == 200:
                    self.log.info('response: ' + str(req.status_code))
                    self.log.info('request text: ' + req.text)
                    raise Exception('non-200 response!')
                transactions = json.loads(req.text).get('transactions', [])
                datum['transactions'] = transactions

            return data

        def add_get_inventoryitems(data, shop, version, req_kwargs):
             # workaround to get inventory item data (i.e. costs) for products
            self.log.info('Requesting inventory items of product variants...')
            base_url = "https://{shop}.myshopify.com/admin/api/{version}/inventory_items.json"
            url = base_url.format(
                shop=shop,
                version=version,
            )

            kwargs = copy.deepcopy(req_kwargs)

            for datum in data:
                ids = [v['inventory_item_id'] for v in datum.get('variants',[])]
                if ids:
                    kwargs['params'] = {'ids': copy.deepcopy(ids)}
                    time.sleep(1) # avoid hitting api call requested limit
                    req = requests.get(url, **kwargs)
                    if not req.status_code == 200:
                        self.log.info('response: ' + str(req.status_code))
                        self.log.info('request text: ' + req.text)
                        raise Exception('non-200 response!')
                    inv_items = json.loads(req.text).get('inventory_items', [])
                    datum['inventory_items'] = inv_items

            return data

        def add_get_events(data, shop, version, req_kwargs):
            # workaround to add events of an order to orders
            self.log.info('Requesting events of orders...')
            base_url = 'https://{shop}.myshopify.com/admin/api/{version}/orders/{id}/events.json'
            base_url = base_url.format(
                shop=shop,
                version=version,
                id='{id}',
            )

            for datum in data:
                id = datum['id']
                time.sleep(1)
                url = base_url.format(id=id)
                req = requests.get(url, **req_kwargs)
                if not req.status_code == 200:
                    self.log.info('response: ' + str(req.status_code))
                    self.log.info('request text: ' + req.text)
                    raise Exception('non-200 response!')
                events = json.loads(req.text).get('events', [])
                datum['events'] = events

            return data


        url = self._base_url.format(**{
            'shop': shop_id,
            'version': self.api_version,
            'object': self.object_metadata.get(
                '_object_url',
                self.shopify_object,
            ),
        })

        # get connection for the applicable shop
        conn = BaseHook.get_connection(source_conn_id)
        login = conn.login
        password = conn.password

        if auth_type == 'access_token':
            headers = {
                'X-Shopify-Access-Token': password,
            }
            kwargs_init = {
                'headers': headers,
                'params': params,
            }
            kwargs_links = {'headers': headers}
        elif auth_type == 'basic_auth':
            kwargs_init = {
                'params': params,
                'auth': HTTPBasicAuth(login, password),
            }
            kwargs_links = {'auth': HTTPBasicAuth(login, password)}
        else:
            raise Exception('Authentication type not accepted!')

        # get and upload data
        self.log.info('Requesting data from REST API - url: {0}, params: {1}' \
            .format(url, str(params)))
        req_kwargs = kwargs_init
        is_first = True
        while is_first or (r.status_code == 200 and url):
            r = requests.get(url, **req_kwargs)
            if is_first:
                is_first = False
                req_kwargs = kwargs_links
            data = json.loads(r.text or '{}').get(self.object_metadata.get(
                '_name_in_request_data',
                self.shopify_object,
            ))
            if self.get_transactions_with_orders:
                data = add_get_transactions(
                    data=data,
                    shop=shop_id,
                    version=self.api_version,
                    req_kwargs=kwargs_links,
                )
            if self.get_events_with_orders:
                data = add_get_events(
                    data=data,
                    shop=shop_id,
                    version=self.api_version,
                    req_kwargs=kwargs_links,
                )
            if self.get_inventory_data_with_product_variants:
                data = add_get_inventoryitems(
                    data=data,
                    shop=shop_id,
                    version=self.api_version,
                    req_kwargs=kwargs_links,
                )
            self.upload_data(data)
            self.log.info('Requesting next page of data...')
            if r.headers.get('Link') and r.headers['Link'][-9:] == 'el="next"':
                url = r.headers['Link'][1:-13]
            else:
                url = None

        if not r.status_code == 200:
            raise Exception('Shopify request returned an error {1}: {0}'.format(
                r.text,
                str(r.status_code),
            ))
